/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ980986960$}}
 */

/*
 * Changes are Copyright (c) 2005 Your Company. All Rights Reserved. 
 * 
 * Generated by an evaluation version of RapidJ for EVALUATION PURPOSES ONLY.
 * 
 * This copyright header can be customised in the licensed version of RapidJ.
 */
package com.company.customer.data.dao;

import java.security.Principal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;

import com.company.customer.data.dso.CustomerDSO;
import com.company.customer.data.to.CustomerTO;
import com.company.customer.util.dao.DAOException;
import com.company.customer.util.dao.DAOUtils;
import com.company.customer.util.dao.PagedData;
import com.company.customer.util.ApplicationException;
import com.company.customer.util.database.DatabaseUtils;
import com.company.customer.util.database.SequenceGenerator;
import com.company.customer.util.database.StatementPreparer;
import com.company.customer.util.database.StatementPreparerException;

/**
 * The CustomerDAO (Customer Data Access Object)
 * class contains methods for retrieving and persisting Customer
 * data.
 *
 * It uses the <a href="http://java.sun.com/blueprints/patterns/TransferObject.html">
 * Transfer Object pattern</a> to return or receive data from a calling component.
 *
 * Data Specification Objects (DSOs) can be used to specifiy what data should
 * be loaded into Transfer Objects.
 */
public class CustomerDAO {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(CustomerDAO.class);

	/**
	 * StatementPreparer for this
	 */
	private StatementPreparer statementPreparer;

	/**
	 * Default constructor.
	 */
	public CustomerDAO() throws StatementPreparerException {
		// Create the StatementPreparer used to create PreparedStatements
		statementPreparer = new StatementPreparer();
		// Add the Customer sql to it
		statementPreparer.addSql("customer-sql.xml", getClass());
	}

	/**
	 * Returns true if the specified Customer exists.
	 *
	 * @param customerId - identifying the Customer.
	 * @return true if the Customer exists.
	 */
	public boolean existsCustomer(
			Long customerId)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;

		// Check the required parameters are present
		if (customerId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Prepare the statement
			Map values = new HashMap();
			values.put("customerId", customerId);
			ps = statementPreparer.prepare(conn, "countCustomer", values);

			// Execute the query
			rs = ps.executeQuery();

			// Get the count
			int count = 0;
			if ((rs!=null) && (rs.next())) {
				count = rs.getInt(1);
			}

			// Process the count
			if (count == 0) {
				// No records
				return false;
			} else {
				return true;
			}

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns a CustomerTO containing data for the specified Customer
	 *
	 * @param customerId - identifying the Customer.
	 * @param customerDSO - specifying what data to return.
	 * @param returnNull - if true and the Customer is not found then null
	 * is returned, else if false and not found then a DAOException is thrown.
	 * @return a populated CustomerTO or null if it did not exist.
	 */
	public CustomerTO getCustomer(
			Long customerId,
			CustomerDSO customerDSO,
			boolean returnNull)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		ArrayList customers = null;

		// Check the required parameters are present
		if (customerId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Prepare the statement
			Map values = new HashMap();
			values.put("customerId", customerId);
			ps = statementPreparer.prepare(conn, "getCustomer", values);

			// Execute the query
			rs = ps.executeQuery();

			// Extract CustomerTO's from the ResultSet
			customers = DAOUtils.extract(CustomerTO.class, rs);

			// Process the results
			if (customers.isEmpty()) {
				// No results, return null or throw exception
				if (returnNull) {
					return null;
				} else {
					throw new DAOException(
						DAOException.SELECT_BY_ID_RETURNED_NO_RECORDS,
						new Object[] {"Customer", customerId}, null);
				}
			} else if (customers.size()==1){
				// Found exactly one record.
				CustomerTO customer = (CustomerTO) customers.get(0);
				// Populate any children according to the DSO
				populateChildren(customer, customerDSO);
				// Return it
				return  customer;
			} else {
				// More than one record is an error
				throw new DAOException(
					DAOException.SELECT_BY_ID_RETURNED_MULTIPLE_RECORDS,
					new Object[] {"Customer", customerId}, null);
			}

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns true if the Customer exists for the specified parent CustomerType
	 *
	 * @param customerId - identifying the Customer.
	 * @param customerTypeId - identifying the parent CustomerType.
	 * @return true if the Customer exists.
	 */
	public boolean existsCustomerForCustomerType(
			Long customerId,
			Long customerTypeId)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;

		// Check the required parameters are present
		if (customerId==null) throw new NullPointerException();
		if (customerTypeId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Prepare the statement
			Map values = new HashMap();
			values.put("customerId", customerId);
			values.put("customerTypeId", customerTypeId);
			ps = statementPreparer.prepare(conn, "countCustomersForCustomerType", values);

			// Execute the query
			rs = ps.executeQuery();

			// Get the count
			int count = 0;
			if ((rs!=null) && (rs.next())) {
				count = rs.getInt(1);
			}

			// Process the count
			if (count == 0) {
				// No records
				return false;
			} else {
				return true;
			}

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns true if any Customer exists for the specified parent CustomerType
	 *
	 * @param customerTypeId - identifying the parent CustomerType.
	 * @return true if any Customer exists for the parent.
	 */
	public boolean existsCustomerForCustomerType(
			Long customerTypeId)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;

		// Check the required parameters are present
		if (customerTypeId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Prepare the statement
			Map values = new HashMap();
			values.put("customerTypeId", customerTypeId);
			ps = statementPreparer.prepare(conn, "countCustomersForCustomerType", values);

			// Execute the query
			rs = ps.executeQuery();

			// Get the count
			int count = 0;
			if ((rs!=null) && (rs.next())) {
				count = rs.getInt(1);
			}

			// Process the count
			if (count == 0) {
				// No records
				return false;
			} else {
				return true;
			}

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns a list of CustomerTO's that are children of the
	 * specified CustomerType.
	 *
	 * @param customerTypeId - identifying the parent customerType.
	 * @param customerDSO - specifying what data should be loaded.
	 */
	public ArrayList getCustomersForCustomerType(
			Long customerTypeId,
			CustomerDSO customerDSO)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		ArrayList customers = null;

		// Check the required parameters are present
		if (customerTypeId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			Map values = new HashMap();
			values.put("customerTypeId", customerTypeId);
			ps = statementPreparer.prepare(conn, "getCustomersForCustomerType", values);

			// Execute the statement
			rs = ps.executeQuery();

			// Extract CustomerTO's from the results
			customers = DAOUtils.extract(CustomerTO.class, rs);

			// Populate children as requested
			populateChildren(customers, customerDSO);

			// Return the list
			return customers;

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns a PagedData of CustomerTO's that are children of the
	 * specified CustomerType.
	 *
	 * @param customerTypeId - identifying the parent customerType.
	 * @param customerDSO - specifying what data should be loaded.
	 * @param pageNumber - the requested page number.
	 * @param pageSize - the maximum number of records per page.
	 */
	public PagedData getCustomersForCustomerType(
			Long customerTypeId,
			CustomerDSO customerDSO,
			int pageNumber,
			int pageSize)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		PagedData customers = null;
		int maxPageNumber = 0;

		// Check the required parameters are present
		if (customerTypeId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			HashMap values = new HashMap();
			values.put("customerTypeId", customerTypeId);
			ps = statementPreparer.prepare(conn, "getCustomersForCustomerType", values);

			// Execute the statement
			rs = ps.executeQuery();

			// Extract CustomerTO's from the results
			customers = DAOUtils.extract(CustomerTO.class, rs, pageNumber, pageSize);

			// Populate children as requested
			populateChildren(customers, customerDSO);

		} catch (DAOException de) {

			if (de.getKey().equals(DAOException.PAGE_DOES_NOT_EXIST)) {
				// Handle this by requesting the max page number.
				maxPageNumber = ((Integer) de.getArgs()[0]).intValue();
				if (logger.isDebugEnabled()) logger.debug("page " + pageNumber + " does not exist, max page is " + maxPageNumber);
			} else {
				// Throw other DAOExceptions up
				logger.error("dao exception", de);
				throw de;
			}

		} finally {

			// Always close the resources
			DAOUtils.close(rs, ps, conn);

		}

		if (customers!=null) {
			// Return the paged date
			return customers;
		} else {
			// The requested page must not exist. Get the last page.
			return getCustomersForCustomerType(
				customerTypeId, customerDSO, maxPageNumber, pageSize);
		}

	}

	/**
	 * Returns a list of CustomerTO's that are not yet associated
	 * with a Address.
	 *
	 * @param customerId - the (optional) Customer to
	 * include in the list regardless.
	 * @param customerDSO - specifying what data should be loaded.
	 */
	public ArrayList getCustomersWithNoAddress(
			Long customerId,
			CustomerDSO customerDSO)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		ArrayList customers = null;

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			HashMap values = new HashMap();
			values.put("customerId", customerId);
			ps = statementPreparer.prepare(conn, "getCustomersWithNoAddress", values);

			// Execute the statement
			rs = ps.executeQuery();

			// Extract CustomerTO's from the results
			customers = DAOUtils.extract(CustomerTO.class, rs);

			// Populate children as requested
			populateChildren(customers, customerDSO);

			// Return the list
			return customers;

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns a list containing all CustomerTO's.
	 *
	 * @param customerDSO - specifying what data should be loaded.
	 */
	public ArrayList getAllCustomers(
			CustomerDSO customerDSO)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		ArrayList customers = null;

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			ps = statementPreparer.prepare(conn, "getAllCustomers");

			// Execute it
			rs = ps.executeQuery();

			// Extract the CustomerTO's
			customers = DAOUtils.extract(CustomerTO.class, rs);

			// Populate children as requested
			populateChildren(customers, customerDSO);

			// Return the list
			return customers;

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

	}

	/**
	 * Returns a PagedData containing CustomerTO's.
	 *
	 * @param customerDSO - specifying what data should be loaded.
	 * @param customerName - optional customer name.
	 * @param pageNumber - the requested page number.
	 * @param pageSize - the maximum number of records per page.
	 * @exception DAOException if there is a problem.
	 */
	public PagedData getAllCustomers(
			CustomerDSO customerDSO,
			String customerName,
			int pageNumber,
			int pageSize)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		PagedData customers = null;
		int maxPageNumber = 0;

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			Map values = new HashMap();
			if (customerName!=null) values.put("customerName", "%"+customerName+"%");
			ps = statementPreparer.prepare(conn, "getAllCustomers", values);

			// Execute it
			rs = ps.executeQuery();

			// Extract the CustomerTO's
			customers = DAOUtils.extract(CustomerTO.class, rs, pageNumber, pageSize);

			// Populate children as requested
			populateChildren(customers, customerDSO);

		} catch (DAOException de) {

			if (de.getKey().equals(DAOException.PAGE_DOES_NOT_EXIST)) {
				// Handle this by requesting the max page number.
				maxPageNumber = ((Integer) de.getArgs()[0]).intValue();
				if (logger.isDebugEnabled()) logger.debug("page " + pageNumber + " does not exist, max page is " + maxPageNumber);
			} else {
				// Throw other DAOExceptions up
				logger.error("dao exception", de);
				throw de;
			}

		} finally {

			// Always close resources
			DAOUtils.close(rs, ps, conn);

		}

		if (customers!=null) {
			// Return the paged date
			return customers;
		} else {
			// The requested page must not exist. Get the last page.
			return getAllCustomers(customerDSO, customerName, maxPageNumber, pageSize);
		}

	}

	/**
	 * Save the specified Customer.
	 *
	 * @param customerTO - containing the data to save.
	 * @param user - the user who is requesting the save.
	 * @return the id of the Customer that was saved.
	 */
	public Long saveCustomer(
			CustomerTO customer,
			Principal user)
			throws ApplicationException,
				SQLException {

		// Determine if the customer is a new one
		boolean isNew = ((customer.getCustomerId()==null) ||
			(customer.getCustomerId().longValue()<=0));

		// Get the current time
		Timestamp now = new Timestamp(System.currentTimeMillis());

		// Get the unique indentifier for the user
		String uid = DAOUtils.getUid(user);

		if (isNew) {
			// Get the next unique id using the key sequence generator
			Long nextId = SequenceGenerator.getNextId("customerId");

			// Setup some initial data values
			customer.setCustomerId(nextId);
			customer.setCreateDatetime(now);
			customer.setCreateUid(uid);
			customer.setVersion(new Long(1));
		}

		// Set the update datetime and updated uid
		customer.setUpdateDatetime(now);
		customer.setUpdateUid(uid);

		// Get the property values from the CustomerTO
		Map values = DAOUtils.getPropertyValues(customer);

		PreparedStatement ps = null;
		Connection conn = null;
		int updateCount;

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			if (isNew) {

				// Insert a new Customer
				// Create the insert statement
				ps = statementPreparer.prepare(conn, "insertCustomer", values);
				// And execute it
				updateCount = ps.executeUpdate();

			} else {

				// Update the existing Customer
				// Create the update statment
				ps = statementPreparer.prepare(conn, "updateCustomer", values);
				// Execute it
				updateCount = ps.executeUpdate();
				if (updateCount==1) {
					// The update was successful.
					logger.debug("update successful");
				} else if (updateCount==0) {
					// No records were affected.
					if (existsCustomer(customer.getCustomerId())) {
						// Record exists, there must have been a concurrent update.
						throw new DAOException(
							DAOException.CONCURRENT_UPDATE,
							new Object[] {"Customer", customer.getCustomerId()}, null);
					} else {
						// Record does not exist.
						throw new DAOException(
							DAOException.RECORD_DOES_NOT_EXIST,
							new Object[] {"Customer", customer.getCustomerId()}, null);
					}
				} else {
					// Multiple (or negative) number of records affected
					logger.debug("updated " + updateCount + " records");
					throw new DAOException(
						DAOException.UPDATE_AFFECTED_WRONG_NUMBER_RECORDS,
						new Object[] {"Customer", customer.getCustomerId()}, null);
				}

			}

		} finally {

			// Always close resources
			DAOUtils.close(ps, conn);

		}

		// Return the id of the Customer that was saved
		return customer.getCustomerId();

	}

	/**
	 * Saves all of the supplied CustomerTO's in the list.
	 *
	 * @param customers the list of CustomerTO's.
	 * @param user - the user who is requesting the save.
	 * @return an array containing the id's of the saved CustomerTO's.
	 */
	public Long[] saveCustomers(
			Collection customers,
			Principal user)
			throws ApplicationException,
				SQLException {

		Long[] customerIds = new Long[customers.size()];
		CustomerTO customer = null;
		int i = 0;
		for (Iterator iter = customers.iterator(); iter.hasNext();) {
			customer = (CustomerTO) iter.next();
			customerIds[i++] = saveCustomer(customer, user);
		}
		return customerIds;
	}

	/**
	 * Deletes the specified Customer
	 *
	 * @param customerId identifying the customerId
	 */
	public void deleteCustomer(
			Long customerId)
			throws ApplicationException,
				SQLException {

		// Check the required parameters are present
		if (customerId==null) throw new NullPointerException();


		// Delete Purchase
		DAOs.getPurchaseDAO().deletePurchasesForCustomer(customerId);

		// Delete CreditCard
		DAOs.getCreditCardDAO().deleteCreditCardsForCustomer(customerId);

		// Delete Address
		DAOs.getAddressDAO().deleteAddressForCustomer(customerId);

		// Delete the Customer
		PreparedStatement ps = null;
		Connection conn = null;

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			HashMap values = new HashMap();
			values.put("customerId", customerId);
			ps = statementPreparer.prepare(conn, "deleteCustomer", values);

			// Get the number of affected records
			int count = ps.executeUpdate();

			// And make sure a single record was deleted
			if (count == 0) {
				throw new DAOException(
					DAOException.DELETE_BY_ID_AFFECTED_NO_RECORDS,
					new Object[] {"Customer", customerId}, null);
			} else if (count != 1) {
				throw new DAOException(
					DAOException.DELETE_BY_ID_AFFECTED_MULTIPLE_RECORDS,
					new Object[] {"Customer", customerId}, null);
			}

		} finally {

			// Always close resources
			DAOUtils.close(ps, conn);

		}

	}

	/**
	 * Deletes all Customers for the specified CustomerType
	 *
	 * @param customerTypeId identifying the CustomerType
	 */
	public void deleteCustomersForCustomerType(
			Long customerTypeId)
			throws ApplicationException,
				SQLException {

		// Get the list of Customers for the specified CustomerType
		ArrayList customers = getCustomersForCustomerType(customerTypeId);

		// And delete them individually
		for (Iterator iter = customers.iterator(); iter.hasNext(); ) {
			CustomerTO customer = (CustomerTO) iter.next();
			deleteCustomer(customer.getCustomerId());
		}

	}

	/**
	 * Clears any references to the specified DefaultCreditCard
	 *
	 * @param defaultCreditCardId identifying the CreditCard
	 */
	public void clearReferencesToDefaultCreditCard(
			Long defaultCreditCardId)
			throws ApplicationException,
				SQLException {

		PreparedStatement ps = null;
		Connection conn = null;

		// Check the required parameters are present
		if (defaultCreditCardId==null) throw new NullPointerException();

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			// Create the statement
			HashMap values = new HashMap();
			values.put("defaultCreditCardId", defaultCreditCardId);
			ps = statementPreparer.prepare(conn, "clearReferencesToDefaultCreditCard", values);

			// Get the number of affected records
			int count = ps.executeUpdate();

		} finally {

			// Always close resources
			DAOUtils.close(ps, conn);

		}

	}

	// ---- Start Convenience Methods ----

	public CustomerTO getCustomer(Long customerId) throws ApplicationException, SQLException {
		return getCustomer(customerId, new CustomerDSO(), false);
	}

	public CustomerTO getCustomer(Long customerId, boolean returnNull) throws ApplicationException, SQLException {
		return getCustomer(customerId, new CustomerDSO(), returnNull);
	}

	public ArrayList getCustomersForCustomerType(Long customerTypeId) throws ApplicationException, SQLException {
		return getCustomersForCustomerType(customerTypeId, new CustomerDSO());
	}

	public PagedData getCustomersForCustomerType(Long customerTypeId, int pageNumber, int pageSize) throws ApplicationException, SQLException {
		return getCustomersForCustomerType(customerTypeId, new CustomerDSO(), pageNumber, pageSize);
	}

	public ArrayList getCustomersWithNoAddress(Long customerId) throws ApplicationException, SQLException {
		return getCustomersWithNoAddress(customerId, new CustomerDSO());
	}

	public ArrayList getAllCustomers() throws ApplicationException, SQLException {
		return getAllCustomers(new CustomerDSO());
	}

	public PagedData getAllCustomers(int pageNumber, int pageSize) throws ApplicationException, SQLException {
		return getAllCustomers(new CustomerDSO(), null, pageNumber, pageSize);
	}

	public PagedData getAllCustomers(String customerName, int pageNumber, int pageSize) throws ApplicationException, SQLException {
		return getAllCustomers(new CustomerDSO(), customerName, pageNumber, pageSize);
	}

	// ---- End Convenience Methods ----

	/**
	 * This method populates the children for each Customer in the specified
	 * list according to the supplied DSO.
	 *
	 * @param customers the list of CustomerTO object to populate
	 * @param customerDSO specifying what children to populate
	 */
	protected void populateChildren(
			ArrayList customers,
			CustomerDSO customerDSO)
			throws ApplicationException,
				SQLException {

		for (Iterator iter = customers.iterator(); iter.hasNext(); ) {
			CustomerTO customer = (CustomerTO) iter.next();
			populateChildren(customer, customerDSO);
		}
	}

	/**
	 * This method populates the children of the specified Customer
	 * according to the supplied DSO.
	 *
	 * @param customer to be populated
	 * @param customerDSO specifies which children to populate
	 */
	protected void populateChildren(
			CustomerTO customer,
			CustomerDSO customerDSO)
			throws ApplicationException,
				SQLException {

		// Purchase
		if (customerDSO.getPurchaseDSO()!=null) {
			customer.setPurchase(DAOs.getPurchaseDAO().
					getPurchasesForCustomer(
						customer.getCustomerId(),
						customerDSO.getPurchaseDSO()));
		}

		// CreditCard
		if (customerDSO.getCreditCardDSO()!=null) {
			customer.setCreditCard(DAOs.getCreditCardDAO().
					getCreditCardsForCustomer(
						customer.getCustomerId(),
						customerDSO.getCreditCardDSO()));
		}

		// Address
		if (customerDSO.getAddressDSO()!=null) {
			customer.setAddress(DAOs.getAddressDAO().
					getAddressForCustomer(
						customer.getCustomerId(),
						customerDSO.getAddressDSO()));
		}

	}

}
