<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=ConvertUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.convert -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.convert;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.apache.struts.util.MessageResources;

/**
 * ConvertUtils provides flexible data conversion functionality through a
 * configurable register of Converters.
 */
public class ConvertUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(ConvertUtils.class);

	/**
	 * Map of Class fromType -> Map of Class toType -> Converter.
	 *
	 * No need to synchronise this because it is only ever modified within
	 * synchronised methods.
	 */
	private Map converters = new HashMap();

	/**
	 * Map of Class fromType -> Map of Class toType -> Converter.
	 *
	 * No need to synchronise this because it is only ever modified within
	 * synchronised methods.
	 */
	private Map equivalentConverters = new HashMap();

	/**
	 * Map of Class fromType -> Set of Class toTypes for which it is know
	 * that there are no equivalent converters available.
	 *
	 * Synchronize the map so it can be safely accessed by multiple threads.
	 */
	private Map noEquivalents = Collections.synchronizedMap(new HashMap());

	/**
	 * The MessageResources that will be available to Converters used by this.
	 */
	private MessageResources resources = null;

	/**
	 * Create an ConvertUtils with no registered Converters.
	 *
	 * @param resources - the MessageResources to use for Converter initialisation.
	 */
	public ConvertUtils(MessageResources resources) {
		this.resources = resources;
	}

	/**
	 * Registers default converters.
	 *
	 * @param resources - the MessageResources to use for Converter initialisation.
	 */
	public static ConvertUtils createDefaultInstance(MessageResources resources) {

		// Create a new empty ConvertUtils
		ConvertUtils cu = new ConvertUtils(resources);

		// Handle booleans
		cu.register(new Converters.BasicBooleanConverter(),
				new Class[] {boolean.class, Boolean.class, String.class},
				new Class[] {boolean.class, Boolean.class, String.class},
				false);

		// Handle bytes
		cu.register(new Converters.BasicByteConverter(),
				new Class[] {byte.class, Byte.class, String.class},
				new Class[] {byte.class, Byte.class, String.class},
				false);

		// Handle shorts
		cu.register(new Converters.BasicShortConverter(),
				new Class[] {short.class, Short.class, String.class},
				new Class[] {short.class, Short.class, String.class},
				false);

		// Handle integers
		cu.register(new Converters.BasicIntegerConverter(),
				new Class[] {int.class, Integer.class, String.class},
				new Class[] {int.class, Integer.class, String.class},
				false);

		// Handle longs
		cu.register(new Converters.BasicLongConverter(),
				new Class[] {long.class, Long.class, String.class},
				new Class[] {long.class, Long.class, String.class},
				false);

		// Handle floats
		cu.register(new Converters.BasicFloatConverter(),
				new Class[] {float.class, Float.class, String.class},
				new Class[] {float.class, Float.class, String.class},
				false);

		// Handle doubles
		cu.register(new Converters.BasicDoubleConverter(),
				new Class[] {double.class, Double.class, String.class},
				new Class[] {double.class, Double.class, String.class},
				false);

		// Handle BigDecimals
		cu.register(new Converters.BasicBigDecimalConverter(),
				new Class[] {BigDecimal.class, String.class},
				new Class[] {BigDecimal.class, String.class},
				false);

		// Handle dates
		cu.register(new Converters.BasicDateConverter(),
				new Class[] {java.sql.Date.class, String.class},
				new Class[] {java.sql.Date.class, String.class},
				false);

		// Handle timestamps
		cu.register(new Converters.BasicTimestampConverter(),
				new Class[] {java.sql.Timestamp.class, String.class},
				new Class[] {java.sql.Timestamp.class, String.class},
				false);

		// Handle time
		cu.register(new Converters.BasicTimeConverter(),
				new Class[] {java.sql.Time.class, String.class},
				new Class[] {java.sql.Time.class, String.class},
				false);

		// Handle strings
		cu.register(new Converters.StringConverter(),String.class, String.class, true);
		cu.register(new Converters.StringConverter(),Object.class, String.class, true);
		cu.register(new Converters.StringConverter(),String.class, Object.class, true);

		// Return the populated ConvertUtils
		return cu;

	}

	/**
	 * Register the converter to convert between fromType to toType.
	 *
	 * @param converter the Converter to register.
	 * @param fromType the type the Converter can convert from.
	 * @param toType the type the Converter can convert to.
	 * @param replace - if true then existing registrations are replaced.
	 */
	public synchronized void register(Converter converter, Class fromType, Class toType, boolean replace) {

		// Initialise the converter
		converter.init(resources);

		// Attempt to locate an existing map of converters that
		// convert from the specified fromType.
		Map map = (Map) converters.get(fromType);
		if (map==null) {
			// No map exists, create one, and register it under fromType
			map = new HashMap();
			converters.put(fromType, map);
		}

		// Check that a converter is not already registered for the specified
		// from and to types.
		if ((map.containsKey(toType)) && (logger.isDebugEnabled())) {
			// A converter is already registered for specified from and to type.)
			logger.debug("a converter is already registered for:");
			logger.debug("fromType=" + fromType);
			logger.debug("toType=" + toType);
			logger.debug("converter already registered=" + map.get(toType));
			logger.debug("registering new converter=" + converter);
		}

		// Register the converter
		map.put(toType, converter);
		if (logger.isDebugEnabled()) logger.debug("registered converter=" +
				converter + ", fromType=" + fromType + ", toType=" + toType);

		// Clear the information that has been cached about which types there
		// are no equivalent converters available for, because with this new
		// addition this info may be incorrect.
		noEquivalents.clear();

	}

	/**
	 * Register the converter to convert between any fromTypes to toType.
	 *
	 * @param converter
	 * @param fromTypes
	 * @param toType
	 * @param replace - if true then existing registrations are replaced.
	 */
	public void register(Converter converter, Class[] fromTypes, Class toType, boolean replace) {
		for (int i = 0; i < fromTypes.length; i++) {
			register(converter, fromTypes[i], toType, replace);
		}
	}

	/**
	 * Register the converter to convert between fromType to any toTypes.
	 *
	 * @param converter
	 * @param fromType
	 * @param toTypes
	 * @param replace - if true then existing registrations are replaced.
	 */
	public void register(Converter converter, Class fromType, Class toTypes[], boolean replace) {
		for (int i = 0; i < toTypes.length; i++) {
			register(converter, fromType, toTypes[i], replace);
		}
	}

	/**
	 * Register the converter to convert between any fromTypes to any toType.
	 *
	 * @param converter
	 * @param fromTypes
	 * @param toTypes
	 * @param replace - if true then existing registrations are replaced.
	 */
	public void register(Converter converter, Class[] fromTypes, Class[] toTypes, boolean replace) {
		for (int i = 0; i < fromTypes.length; i++) {
			for (int j = 0; j < toTypes.length; j++) {
				register(converter, fromTypes[i], toTypes[j], replace);
			}
		}
	}

	/**
	 * Returns true if ConvertUtils can convert directly between the
	 * specified types.
	 *
	 * @param fromType
	 * @param toType
	 * @return
	 */
	public boolean canConvert(Class fromType, Class toType) {
		return (getConverter(fromType, toType, true)!=null);
	}

	/**
	 * Converts the specified value from type fromClass to toType for the specified
	 * locale.
	 *
	 * @param value - the value to convert.
	 * @param fromType - the current type of the value.
	 * @param toType - the type to convert the value to.
	 * @param locale - the locale under which to make the conversion.
	 * @return the converted object.
	 */
	public Object convert(Object value, Class fromType, Class toType, Locale locale) {
		Converter converter = getConverter(fromType, toType, true);
		if (converter!=null) {
			return converter.convert(value, fromType, toType, locale);
		} else {
			throw new RuntimeException("No converter!");
		}
	}

	/**
	 * Returns a Converter that can convert directly between the specified
	 * types. The returned converter may be an equivalent for the specified
	 * from and to type.
	 *
	 * @param fromType - the type the Converter should be able to convert from.
	 * @param toType - the type the Converter should be able to convert to.
	 * @return
	 */
	public Converter getConverter(Class fromType, Class toType) {
		return getConverter(fromType, toType, true);
	}

	/**
	 * Returns a Converter that can convert directly between the specified types.
	 *
	 * @param fromType - the type the Converter should be able to convert from.
	 * @param toType - the type the Converter should be able to convert to.
	 * @param equivalentOk - if true then if a converter is not found for the
	 * specified from and to types, then the converters are searched for an
	 * equivalent converter based on subclasses/subinterfaces of toType and
	 * superclasses/superinterfaces of fromType.
	 * @return the Converter or null if none found.
	 */
	public synchronized Converter getConverter(Class fromType, Class toType, boolean equivalentOk) {

		Converter converter = null;

		// Get the registered converter for the specified types.
		Map map = (Map) converters.get(fromType);
		if (map!=null) {
			converter = (Converter) map.get(toType);
		}

		// If a Converter was found, or even if one was not but equivalentOk is
		// false, return whatever was found.
		if ((converter!=null) || (!equivalentOk)) {
			return converter;
		}

		// Search for an equivalent.

		// See if it has already been registered in the equivalent converters.
		map = (Map) equivalentConverters.get(fromType);
		if (map!=null) {
			converter = (Converter) map.get(toType);
			if (converter!=null) {
				if (logger.isDebugEnabled()) logger.debug("found cached equivalent converter");
				return converter;
			}
		}

		// So it's not already registered in the equivalent converters.
		// Check the known list of previously requested fromType/toType
		// combinations that are known not to have a registered converter.
		if (isNoEquivalentAvailable(fromType,toType)) {
			if (logger.isDebugEnabled()) logger.debug("no equivalent is available");
			return null;
		}

		// Look for converters that have been registered under sub classes/
		// interfaces of toType and/or super classes/interfaces of fromType.

		Class fromTypeOther = null;
		Class toTypeOther = null;
		Map toTypes = null;

		// Iterate through fromTypes determined first from the original fromType,
		// then from implemented interfaces then from super classes.
		ArrayList fromTypes = new ArrayList();
		// Start with the original fromType
		fromTypes.add(fromType);
		// Then try the interfaces
		Class[] fromTypeInterfaces = fromType.getInterfaces();
		for (int i = 0; i < fromTypeInterfaces.length; i++ ) {
			fromTypes.add(fromTypeInterfaces[i]);
		}
		// Then add super classes (including Object)
		fromTypeOther = fromType;
		if (!fromType.equals(Object.class)) {
			do {
				fromTypeOther = fromTypeOther.getSuperclass();
				fromTypes.add(fromTypeOther);
			} while (!fromTypeOther.equals(Object.class));
		}

		// Iterate through the fromTypes
		outer: for (Iterator iter = fromTypes.iterator(); iter.hasNext(); ) {

			fromTypeOther = (Class) iter.next();
			toTypes = (Map) converters.get(fromTypeOther);

			if (toTypes!=null) {
				// See if a conversion exists to the original toType.
				// (Note: This will cause a duplicate check when
				// fromTypeOther is the original fromType, but thats ok)
				converter = (Converter) toTypes.get(toType);
				if (converter!=null) {
					// Found one! Set toTypeOther so the equiv converter
					// can be created correctly..
					toTypeOther = toType;
					break outer;
				}
				// Try and locate a converter for subclass/subinterface
				// of toType, and we can use this..
				for (Iterator iter2 = toTypes.keySet().iterator(); iter2.hasNext(); ) {
					toTypeOther = (Class) iter2.next();
					if (!(toTypeOther.equals(toType)) &&
						(toType.isAssignableFrom(toTypeOther))) {
						// We have found a converter that we can use.
						converter = (Converter) toTypes.get(toTypeOther);
						break outer;
					}
				}
			}
		}

		// Debug
		if (logger.isDebugEnabled()) {
			if (converter!=null) {
				logger.debug("found alternative converter for:");
				logger.debug("toType=" + toType);
				logger.debug("fromType=" + fromType);
				logger.debug("using converter registered for:");
				logger.debug("toTypeOther=" + toTypeOther);
				logger.debug("fromTypeOther=" + fromTypeOther);
				logger.debug("converter=" + converter);
			} else {
				logger.debug("did not find alternative converter for:");
				logger.debug("toType=" + toType);
				logger.debug("fromType=" + fromType);
			}
		}

		if (converter==null) {
			// If converter was no found, regster that there is no
			// equivalent available for the specified from and to types
			setNoEquivalentAvailable(fromType, toType);
		} else {
			// Else add the converter to the equivalent converters map.
			map = (Map) equivalentConverters.get(fromType);
			if (map==null) {
				// No map exists, create one, and register it under fromType
				map = new HashMap();
				equivalentConverters.put(fromType, map);
			}
			// Create and register the equivalent converter
			converter = new EquivalentConverter(converter, fromTypeOther, toTypeOther);
			map.put(toType, converter);
		}

		// Return whatever we have found (maybe nothing)
		return converter;

	}

	/**
	 * Debugs out the internal state
	 */
	public void debug() {
		if (logger.isDebugEnabled()) {
			logger.debug("-------------- ConvertUtils Internal State ---------------");

			logger.debug("number registered type conversions=" + count(converters));
			logger.debug("number equivalent type conversions=" + count(equivalentConverters));
			logger.debug("known number of type conversions with no converter=" + count(noEquivalents));
			logger.debug("registered converters:");
			for (Iterator iter = converters.keySet().iterator(); iter.hasNext(); ) {
				Class fromType = (Class) iter.next();
				logger.debug("  from: " + fromType.getName());
				Map toTypes = (Map) converters.get(fromType);
				for (Iterator iter2 = toTypes.keySet().iterator(); iter2.hasNext(); ) {
					Class toType = (Class) iter2.next();
					logger.debug("    to: " + toType.getName());
					logger.debug("        " + toTypes.get(toType));
				}
			}
			logger.debug("equivalent converters:");
			for (Iterator iter = equivalentConverters.keySet().iterator(); iter.hasNext(); ) {
				Class fromType = (Class) iter.next();
				logger.debug("  from: " + fromType.getName());
				Map toTypes = (Map) equivalentConverters.get(fromType);
				for (Iterator iter2 = toTypes.keySet().iterator(); iter2.hasNext(); ) {
					Class toType = (Class) iter2.next();
					logger.debug("    to: " + toType.getName());
					logger.debug("        " + toTypes.get(toType));
				}
			}
			logger.debug("type conversions having no registered converters:");
			for (Iterator iter = noEquivalents.keySet().iterator(); iter.hasNext(); ) {
				Class fromType = (Class) iter.next();
				logger.debug("  from: " + fromType.getName());
				Set toTypes = (Set) noEquivalents.get(fromType);
				for (Iterator iter2 = toTypes.iterator(); iter2.hasNext(); ) {
					Class toType = (Class) iter2.next();
					logger.debug("    to: " + toType.getName());
				}
			}
			logger.debug("----------------------------------------------------------");
		}
	}

	/**
	 * Utility method to count the total number of records in a Map
	 * containing Collection or Map objects.
	 *
	 * @param map
	 * @return
	 */
	private static int count(Map map) {
		Object subObject = null;
		int count = 0;
		for (Iterator iter = map.values().iterator(); iter.hasNext(); ) {
			subObject = iter.next();
			if (subObject instanceof Collection) {
				count += ((Collection)subObject).size();
			} else if (subObject instanceof Map) {
				count += ((Map)subObject).size();
			} else {
				throw new RuntimeException("Sub object not Collection or Map. subObject=" + subObject);
			}
		}
		return count;
	}

	/**
	 * Returns true if it is known that there is no equivalent converter
	 * available for the specified from and to types.
	 *
	 * @param fromType
	 * @param toType
	 * @return
	 */
	private boolean isNoEquivalentAvailable(Class fromType, Class toType) {
		Set set = (Set) noEquivalents.get(fromType);
		if (set!=null) {
			return set.contains(toType);
		}
		// We do not know that no equivalent is available
		return false;
	}

	/**
	 * Register that there is no equivalent Converter available for the
	 * specified from and to types.
	 *
	 * @param fromType
	 * @param toType
	 * @return
	 */
	private void setNoEquivalentAvailable(Class fromType, Class toType) {
		Set set = (Set) noEquivalents.get(fromType);
		if (set==null) {
			set = new HashSet();
			noEquivalents.put(fromType, set);
		}
		set.add(toType);
	}

	/**
	 * Used to store the original types that an equivalent
	 * Converter was registered for.
	 */
	private static class EquivalentConverter implements Converter {

		private Converter converter;
		private Class originalFromType;
		private Class originalToType;

		/**
		 * Construct a new EquivalentConveter that uses the supplied
		 * converter that was originally registered for the specified
		 * to and from types.
		 *
		 * @param converter
		 * @param originalFromType
		 * @param originalToType
		 */
		public EquivalentConverter(Converter converter, Class originalFromType, Class originalToType) {
			this.converter = converter;
			this.originalFromType = originalFromType;
			this.originalToType = originalToType;
		}

		/**
		 * Initialise this.
		 */
		public void init(MessageResources resources) {
			// Nothing to do because converter has already been initialised.
		}

		/**
		 * Convert the value using the wrapped converter and using the original
		 * from and to types.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Delegate to the wrapped Converter.
			return converter.convert(value, originalFromType, originalToType, locale);
		}

		/**
		 * toString impl.
		 */
		public String toString() {
			StringBuffer buf = new StringBuffer();
			buf.append("[EquivalentConverter");
			buf.append(" converter=").append(converter);
			buf.append(", originalFromType=").append(originalFromType);
			buf.append(", originalToType=").append(originalToType);
			buf.append("]");
			return buf.toString();
		}

	}

}
