<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=Converters.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.convert -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.convert;

import java.math.BigDecimal;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.beanutils.ConversionException;
import org.apache.struts.util.MessageResources;

/**
 * A bunch of basic Converters for the main Java types.
 */
public class Converters {

	// --------------- Message Resource Keys ------------------ //

	/**
	 * Key used to store the default format string for java.lang.Byte in resources.
	 */
	public static final String BASICBYTECONVERTER_FORMATKEY = "basicByteConverter.formatKey";

	/**
	 * Key used to store the default format string for java.lang.Short in resources.
	 */
	public static final String BASICSHORTCONVERTER_FORMATKEY = "basicShortConverter.formatKey";

	/**
	 * Key used to store the default format string for java.lang.Integer in resources.
	 */
	public static final String BASICINTEGERCONVERTER_FORMATKEY = "basicIntegerConverter.formatKey";

	/**
	 * Key used to store the default format string for java.lang.Long in resources.
	 */
	public static final String BASICLONGCONVERTER_FORMATKEY = "basicLongConverter.formatKey";

	/**
	 * Key used to store the default format string for java.lang.Float in resources.
	 */
	public static final String BASICFLOATCONVERTER_FORMATKEY = "basicFloatConverter.formatKey";

	/**
	 * Key used to store the default format string for java.lang.Double in resources.
	 */
	public static final String BASICDOUBLECONVERTER_FORMATKEY = "basicDoubleConverter.formatKey";

	/**
	 * Key used to store the default format string for Date in resources.
	 */
	public static final String BASICDATECONVERTER_FORMATKEY = "basicDateConverter.formatKey";

	/**
	 * Key used to store the default format string for Timestamp in resources.
	 */
	public static final String BASICTIMESTAMPCONVERTER_FORMATKEY = "basicTimestampConverter.formatKey";

	/**
	 * Key used to store the default format string for Time in resources.
	 */
	public static final String BASICTIMECONVERTER_FORMATKEY = "basicTimeConverter.formatKey";

	// --------------- Converter Implementations -------------- //

	/**
	 * Basic Byte Converter.
	 */
	public static class BasicByteConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map byteFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Byte.class, byte.class
		 * to any of:
		 *     String.class, Byte.class, byte.class
		 *
		 * While ensuring that the the data represents a valid Byte.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Byte based on fromType
			Byte byteObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getByteFormat(locale);
							if (f!=null) {
								byteObj = new Byte(f.parse((String)value).byteValue());
							} else {
								byteObj = Byte.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Byte.class)) || (fromType.equals(byte.class))) {
					byteObj = (Byte) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Byte to required toType
			if (byteObj==null) {
				 if (toType.equals(byte.class)) {
					return new Byte((byte)0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Byte.class)) || (toType.equals(byte.class))) {
				return byteObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getByteFormat(locale);
				if (f!=null) {
					return f.format(byteObj);
				} else {
					return byteObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format bytes.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getByteFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICBYTECONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) byteFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					byteFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic Short Converter.
	 */
	public static class BasicShortConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map shortFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Short.class, short.class
		 * to any of:
		 *     String.class, Short.class, short.class
		 *
		 * While ensuring that the the data represents a valid Short.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Short based on fromType
			Short shortObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getShortFormat(locale);
							if (f!=null) {
								shortObj = new Short(f.parse((String)value).shortValue());
							} else {
								shortObj = Short.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Short.class)) || (fromType.equals(short.class))) {
					shortObj = (Short) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Short to required toType
			if (shortObj==null) {
				 if (toType.equals(short.class)) {
					return new Short((short)0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Short.class)) || (toType.equals(short.class))) {
				return shortObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getShortFormat(locale);
				if (f!=null) {
					return f.format(shortObj);
				} else {
					return shortObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format shorts.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getShortFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICSHORTCONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) shortFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					shortFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic Integer Converter.
	 */
	public static class BasicIntegerConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map integerFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Integer.class, integer.class
		 * to any of:
		 *     String.class, Integer.class, integer.class
		 *
		 * While ensuring that the the data represents a valid Integer.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Integer based on fromType
			Integer integerObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getIntegerFormat(locale);
							if (f!=null) {
								integerObj = new Integer(f.parse((String)value).intValue());
							} else {
								integerObj = Integer.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Integer.class)) || (fromType.equals(int.class))) {
					integerObj = (Integer) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Integer to required toType
			if (integerObj==null) {
				 if (toType.equals(int.class)) {
					return new Integer(0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Integer.class)) || (toType.equals(int.class))) {
				return integerObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getIntegerFormat(locale);
				if (f!=null) {
					return f.format(integerObj);
				} else {
					return integerObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format integers.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getIntegerFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICINTEGERCONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) integerFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					integerFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic Long Converter.
	 */
	public static class BasicLongConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map longFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Long.class, long.class
		 * to any of:
		 *     String.class, Long.class, long.class
		 *
		 * While ensuring that the the data represents a valid Long.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Long based on fromType
			Long longObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getLongFormat(locale);
							if (f!=null) {
								longObj = new Long(f.parse((String)value).longValue());
							} else {
								longObj = Long.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Long.class)) || (fromType.equals(long.class))) {
					longObj = (Long) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Long to required toType
			if (longObj==null) {
				 if (toType.equals(long.class)) {
					return new Long((long)0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Long.class)) || (toType.equals(long.class))) {
				return longObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getLongFormat(locale);
				if (f!=null) {
					return f.format(longObj);
				} else {
					return longObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format longs.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getLongFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICLONGCONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) longFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					longFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic Float Converter.
	 */
	public static class BasicFloatConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map floatFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Float.class, float.class
		 * to any of:
		 *     String.class, Float.class, float.class
		 *
		 * While ensuring that the the data represents a valid Float.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Float based on fromType
			Float floatObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getFloatFormat(locale);
							if (f!=null) {
								floatObj = new Float(f.parse((String)value).floatValue());
							} else {
								floatObj = Float.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Float.class)) || (fromType.equals(float.class))) {
					floatObj = (Float) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Float to required toType
			if (floatObj==null) {
				 if (toType.equals(float.class)) {
					return new Float((float)0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Float.class)) || (toType.equals(float.class))) {
				return floatObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getFloatFormat(locale);
				if (f!=null) {
					return f.format(floatObj);
				} else {
					return floatObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format floats.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getFloatFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICFLOATCONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) floatFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					floatFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic Double Converter.
	 */
	public static class BasicDoubleConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DecimalFormat)
		 */
		private Map doubleFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     String.class, Double.class, double.class
		 * to any of:
		 *     String.class, Double.class, double.class
		 *
		 * While ensuring that the the data represents a valid Double.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a Double based on fromType
			Double doubleObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DecimalFormat f = getDoubleFormat(locale);
							if (f!=null) {
								doubleObj = new Double(f.parse((String)value).doubleValue());
							} else {
								doubleObj = Double.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if ((fromType.equals(Double.class)) || (fromType.equals(double.class))) {
					doubleObj = (Double) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Double to required toType
			if (doubleObj==null) {
				 if (toType.equals(double.class)) {
					return new Double((double)0);
				 } else {
					return null;
				 }
			} else if ((toType.equals(Double.class)) || (toType.equals(double.class))) {
				return doubleObj;
			} else if (toType.equals(String.class)) {
				DecimalFormat f = getDoubleFormat(locale);
				if (f!=null) {
					return f.format(doubleObj);
				} else {
					return doubleObj.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DecimalFormat object that can be used
		 * to parse and format doubles.
		 *
		 * @param locale
		 * @return
		 */
		private DecimalFormat getDoubleFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICDOUBLECONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DecimalFormat for this format
				DecimalFormat bf = (DecimalFormat) doubleFormats.get(format);
				if (bf==null) {
					// Has not been created yet. Create one and cache it.
					bf = new DecimalFormat(format);
					doubleFormats.put(format, bf);
				}
				return bf;
			}
		}

	}

	/**
	 * Basic BigDecimal Converter.
	 */
	public static class BasicBigDecimalConverter extends AbstractConverter {

		/**
		 * Converts from any of:
		 *     String.class, BigDecimal.class
		 * to any of:
		 *     String.class, BigDecimal.class
		 *
		 * While ensuring that the the data represents a valid BigDecimal.
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			// Convert the value to a BigDecimal based on fromType
			BigDecimal bigDecimalObj = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							bigDecimalObj = new BigDecimal((String)value);
						} catch (Exception e) {
							throw new ConversionException(e);
						}
					}
				} else if (fromType.equals(BigDecimal.class)) {
					bigDecimalObj = (BigDecimal) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the BigDecimal to required toType
			if (bigDecimalObj==null) {
				return null;
			} else if (toType.equals(BigDecimal.class)) {
				return bigDecimalObj;
			} else if (toType.equals(String.class)) {
				return bigDecimalObj.toString();
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

	}

	/**
	 * String Converter.
	 */
	public static class StringConverter implements Converter {

		/**
		 * Converts from an Object to a String
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {
			if (value==null) {
				return null;
			} else {
				return value.toString();
			}
		}

		/**
		 * Initialise this
		 */
		public void init(MessageResources resources) {
			// Nothing to do
		}

	}

	/**
	 * Basic Boolean Converter.
	 */
	public static class BasicBooleanConverter implements Converter {

		/**
		 * Converts from any of:
		 *     Boolean.class, boolean.class, String.class
		 * to any of
		 *     Boolean.class, boolean.class, String.class
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {

			// Convert the value to a date based on fromType
			Boolean bool = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					bool = new Boolean((String)value);
				} else if ((fromType.equals(Boolean.class)) || (fromType.equals(boolean.class))) {
					bool = (Boolean) value;
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Boolean to required toType
			if (bool==null) {
				if (toType.equals(boolean.class)) {
					return Boolean.FALSE;
				} else {
					return null;
				}
			} else if ((toType.equals(Boolean.class)) || (toType.equals(boolean.class))) {
				return bool;
			} else if (toType.equals(String.class)) {
				return bool.toString();
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Initialise this
		 */
		public void init(MessageResources resources) {
			// Nothing to do
		}

	}

	/**
	 * Basic Date Converter.
	 */
	public static class BasicDateConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DateFormat)
		 */
		private Map dateFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     Date, String
		 * to any of
		 *     Date, String
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {

			// Convert the value to a date based on fromType
			Date date = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DateFormat f = getDateFormat(locale);
							if (f!=null) {
								date = new Date(f.parse((String)value).getTime());
							} else {
								date = Date.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e.getMessage());
						}
					}
				} else if (fromType.equals(Date.class)) {
					date = (Date) ((Date) value).clone();
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Date to required toType
			if (date==null) {
				return null;
			} else if (toType.equals(Date.class)) {
				return date;
			} else if (toType.equals(String.class)) {
				DateFormat f = getDateFormat(locale);
				if (f!=null) {
					return f.format(date);
				} else {
					return date.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DateFormat object that can be used to parse
		 * and format dates.
		 *
		 * @param locale
		 * @return
		 */
		private DateFormat getDateFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICDATECONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DateFormat for this format
				DateFormat f = (DateFormat) dateFormats.get(format);
				if (f==null) {
					// Has not been created yet. Create one and cache it.
					f = new SimpleDateFormat(format);
					dateFormats.put(format, f);
				}
				return f;
			}
		}

	}

	/**
	 * Basic Timestamp Converter.
	 */
	public static class BasicTimestampConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DateFormat)
		 */
		private Map timestampFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     Timestamp, String
		 * to any of
		 *     Timestamp, String
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {

			// Convert the value to a timestamp based on fromType
			Timestamp timestamp = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DateFormat f = getTimestampFormat(locale);
							if (f!=null) {
								timestamp = new Timestamp(f.parse((String)value).getTime());
							} else {
								timestamp = Timestamp.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e.getMessage());
						}
					}
				} else if (fromType.equals(Timestamp.class)) {
					timestamp = (Timestamp) ((Timestamp) value).clone();
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Timestamp to required toType
			if (timestamp==null) {
				return null;
			} else if (toType.equals(Timestamp.class)) {
				return timestamp;
			} else if (toType.equals(String.class)) {
				DateFormat f = getTimestampFormat(locale);
				if (f!=null) {
					return f.format(timestamp);
				} else {
					return timestamp.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DateFormat object that can be used to parse
		 * and format timestamps.
		 *
		 * @param locale
		 * @return
		 */
		private DateFormat getTimestampFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICTIMESTAMPCONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DateFormat for this format
				DateFormat f = (DateFormat) timestampFormats.get(format);
				if (f==null) {
					// Has not been created yet. Create one and cache it.
					f = new SimpleDateFormat(format);
					timestampFormats.put(format, f);
				}
				return f;
			}
		}

	}

	/**
	 * Basic Time Converter.
	 */
	public static class BasicTimeConverter extends AbstractConverter {

		/**
		 * Map of (String) format -> (DateFormat)
		 */
		private Map timeFormats = new HashMap();

		/**
		 * Converts from any of:
		 *     Time, String
		 * to any of
		 *     Time, String
		 */
		public Object convert(Object value, Class fromType, Class toType, Locale locale) {

			// Convert the value to a time based on fromType
			Time time = null;
			if (value!=null) {
				if (fromType.equals(String.class)) {
					if (!value.equals("")) {
						try {
							DateFormat f = getTimeFormat(locale);
							if (f!=null) {
								time = new Time(f.parse((String)value).getTime());
							} else {
								time = Time.valueOf((String)value);
							}
						} catch (Exception e) {
							throw new ConversionException(e.getMessage());
						}
					}
				} else if (fromType.equals(Time.class)) {
					time = (Time) ((Time) value).clone();
				} else {
					throw new RuntimeException("Unhandled fromType=" + fromType);
				}
			}
			// Convert the Time to required toType
			if (time==null) {
				return null;
			} else if (toType.equals(Time.class)) {
				return time;
			} else if (toType.equals(String.class)) {
				DateFormat f = getTimeFormat(locale);
				if (f!=null) {
					return f.format(time);
				} else {
					return time.toString();
				}
			} else {
				throw new RuntimeException("Unhandled toType=" + toType);
			}
		}

		/**
		 * Returns a lazy loaded DateFormat object that can be used to parse
		 * and format times.
		 *
		 * @param locale
		 * @return
		 */
		private DateFormat getTimeFormat(Locale locale) {
			if (resources==null) {
				// No resources so cannot create Format object
				return null;
			}
			String format = resources.getMessage(locale, BASICTIMECONVERTER_FORMATKEY);
			if (format==null) {
				// No format for this
				return null;
			} else {
				// Try get the cached DateFormat for this format
				DateFormat f = (DateFormat) timeFormats.get(format);
				if (f==null) {
					// Has not been created yet. Create one and cache it.
					f = new SimpleDateFormat(format);
					timeFormats.put(format, f);
				}
				return f;
			}
		}

	}

}
