<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=CopyUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.copy -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.copy;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;

import ${project.rootPackage}.util.convert.ConvertUtils;

/**
 * CopyUtils provides functionality to copy data between similar objects.
 * Different data types can be copied through the use of ConvertUtils.
 */
public class CopyUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(CopyUtils.class);

	/**
	 * Map containing cached ClassInfo objects. Class -> ClassInfo.
	 */
	private static final Map classInfos = new HashMap();

	/**
	 * Empty Object Array
	 */
	private static Object[] EMPTY_OBJECT_ARRAY = new Object[] {};

	/**
	 * <p>Copies property values from the source bean to the destination bean
	 * using standard getter and setter methods and converting where necessary
	 * and possible.
	 *
	 * <p>The following scenario, copying from objectA (source) to objectB
	 * (destination) illustrates how the copy works.
     *
     * <ul>
     *   <li>If either objectA or objectB is null then nothing can be done.</li>
     *   <li>Else each getter on objectA is examined in turn:
     *     <ul>
     *       <li>If there exists a corresponding setter method on objectB and
     *           there exists a Converter than converts between the getter and
     *           setter types then the conversion is made and the resulting
     *           value is set on objectB.</li>
     *       <li>Else if the value returned by the getter on objectA is null
     *           then null is set on the dest object if possible.</li>
     *       <li>Else if the value returned by the corresponding getter on
     *           objectB is not null then:
	 *         <ul>
	 *           <li>If both values are collections:
	 *             <ul>
	 *               <li>If the source collection is empty then clear the dest
	 *                   collection.</li>
	 *               <li>Else if the source collection contains objects that
	 *                   can be converted to to themselves then clear the dest
	 *                   collection and add the converted objects</li>
	 *               <li>Else if collections are the same size then copy
	 *                   objects from the source collection to the destination
	 *                   collection using this copy method.</li>
	 *             </ul>
	 *           </li>
	 *           <li> Else do a recursive copy from the object returned by
	 *                objectA to the object returned by objectB.
	 *           </li>
	 *         </ul>
	 *       </li>
	 *     </ul>
	 *   </li>
	 * </ul>
	 *
	 * @param dest - the object to copy values into.
	 * @param source - the object to copy values from.
	 * @param convertUtils - used to convert values.
	 * @param locale - the current locale.
	 * @exception CopyException - thrown when there is a problem getting or
	 * setting values.
	 */
	public static void copy(Object dest,
							Object source,
							ConvertUtils convertUtils,
							Locale locale)
							throws CopyException {

		if (logger.isDebugEnabled()) {
			logger.debug("method called..");
			logger.debug("dest=" + dest);
			logger.debug("source=" + source);
		}

		// If either destination or source is null, there is nothing to do.
		if ((dest==null) || (source==null)) {
			return;
		}

		// Gather some information about the source and dest objects.
		Class sourceClass = source.getClass();
		ClassInfo sourceClassInfo = getClassInfo(sourceClass);
		Class destClass = dest.getClass();
		ClassInfo destClassInfo = getClassInfo(destClass);

		// Iterate through all the methods on the source object and process
		// the getter methods according to the rules.
		Iterator attributeNames = sourceClassInfo.getGetterMethods().keySet().iterator();
		while (attributeNames.hasNext()) {

			// Get the source getter method and get the value from it
			String attributeName = (String) attributeNames.next();
			Method sourceGetterMethod = sourceClassInfo.getGetterMethod(attributeName);
			Class sourceGetterType = sourceGetterMethod.getReturnType();
			Object sourceValue = invoke(sourceGetterMethod, source, EMPTY_OBJECT_ARRAY);

			Method destSetterMethod = destClassInfo.getSetterMethod(attributeName);
			Class destSetterType = null;

			if (logger.isDebugEnabled()) {
				logger.debug("---------- Processing >>" + sourceGetterMethod.getName() + "<< ----------");
				logger.debug("sourceGetterMethod=" + sourceGetterMethod);
				logger.debug("sourceValue=" + sourceValue);
				logger.debug("destSetterMethod=" + destSetterMethod);
			}

			if (destSetterMethod!=null) {

				// There is a corresponding setter method.
				destSetterType = destSetterMethod.getParameterTypes()[0];

				// See if ConvertUtils can convert between the source
				// getter type and dest setter type.
				if (convertUtils.canConvert(sourceGetterType, destSetterType)) {
					Object convertedValue = convertUtils.convert(
							sourceValue, sourceGetterType, destSetterType, locale);
					if (logger.isDebugEnabled()) {
						logger.debug("setting convertedValue=" + convertedValue);
					}
					invoke(destSetterMethod, dest, new Object[] {convertedValue});
					continue;
				}

				// Otherwise, if the source value is null, set null on the dest setter
				if (sourceValue==null) {
					if (logger.isDebugEnabled()) {
						logger.debug("setting null");
					}
					invoke(destSetterMethod, dest, new Object[] {null});
					continue;
				}

			}

			Method destGetterMethod = destClassInfo.getGetterMethod(attributeName);
			if (logger.isDebugEnabled()) {
				logger.debug("destGetterMethod=" + destGetterMethod);
			}
			if (destGetterMethod!=null) {

				Object destValue = invoke(destGetterMethod, dest, EMPTY_OBJECT_ARRAY);
				if (logger.isDebugEnabled()) logger.debug("destValue=" + destValue);

				if (destValue != null) {

					Class destGetterType = destGetterMethod.getReturnType();
					if (Collection.class.isAssignableFrom(destGetterType) &&
							Collection.class.isAssignableFrom(sourceGetterType)) {

						// Both collections
						Collection sourceCollection = (Collection) sourceValue;
						Collection destCollection = (Collection) destValue;

						if (sourceCollection==null) {
							logger.debug("null source collection, but no corresponding setter so ignoring");
						} else if (sourceCollection.isEmpty()) {
							logger.debug("empty source collection, clearing dest collection");
							destCollection.clear();
						} else {
							// Determine the type of the first object in the source collection.
							// Assume that the rest of the objects are of the same type.
							Iterator sourceColValues = sourceCollection.iterator();
							Object sourceColValue = sourceColValues.next();
							Class sourceColValueClass = sourceColValue.getClass();

							if (convertUtils.canConvert(sourceColValueClass, sourceColValueClass)) {
								// Clear dest collection and copy converted values across
								destCollection.clear();
								if (logger.isDebugEnabled()) {
									logger.debug("cleared destCollection");
								}
								Object convertedValue = convertUtils.convert(sourceColValue,
										sourceColValueClass, sourceColValueClass, locale);
								destCollection.add(convertedValue);
								while (sourceColValues.hasNext()) {
									sourceColValue = sourceColValues.next();
									convertedValue = convertUtils.convert(sourceColValue,
											sourceColValueClass, sourceColValueClass, locale);
									destCollection.add(convertedValue);
								}

							} else if (sourceCollection.size()==destCollection.size()) {
								// Copy objects from the source collection to the destination collection
								Iterator destColValues = destCollection.iterator();
								Object destColValue = destColValues.next();
								do {

									copy(destColValue, sourceColValue, convertUtils, locale);

									if (sourceColValues.hasNext()) {
										sourceColValue = sourceColValues.next();
										destColValue = destColValues.next();
									} else {
										break;
									}

								} while(true);

							} else {
								logger.warn("cannot copy non convertible objects bewteen different sized collections");
							}

						}

					} else {

						// Copy the object returned by the source getter method
						// to the object returned by the dest getter method.
						if (logger.isDebugEnabled()) {
							logger.debug("recursive call to copy");
						}
						copy(destValue, sourceValue, convertUtils, locale);

					}

				}

			} // end while (attributeNames.hasNext()) {

		}

	}

	/**
	 * Invokes the method on the object using the params and returns the result.
	 *
	 * @param method - the method to be invoked
	 * @param object - the object to invoke the method on
	 * @param params - the parameters to invoke the method with
	 * @exception CopyException if there is an Exception invoking the method
	 * @return the return from the method invocation
	 */
	private static Object invoke(Method method, Object object, Object[] params) throws CopyException {

		try {

			// Invoke the method
			return method.invoke(object,params);

		} catch (Exception e) {

			// Log some information
			logger.error("exception invoking method");
			logger.error("method=" + method.getName());
			logger.error("object.getClass()=" + object.getClass());
			logger.error("object=" + object);
			logger.error("params=" + Arrays.asList(params));
			logger.error("exception was: " + e.getMessage(), e);
			// Wrap with a CopyException
			throw new CopyException(CopyException.COPY_ERROR, null, e);

		}

	}

	/**
	 * Returns a cached ClassInfo object for the specified class.
	 *
	 * @param clazz
	 * @return
	 */
	private static ClassInfo getClassInfo(Class clazz) {

		// See if there is one cached
		ClassInfo classInfo = (ClassInfo) classInfos.get(clazz);
		if (classInfo==null) {
			// Nothing in the cache for this class, create and cache it.
			classInfo = new ClassInfo(clazz);
			classInfos.put(clazz, classInfo);
		}
		return classInfo;

	}

	/**
	 * Provides fast access to getter and setter methods for a Java Class.
	 */
	private static class ClassInfo {

		/**
		 * The class that the info is for
		 */
		private Class clazz;

		/**
		 * Map of (String) attribute name -> (Method) setter method
		 */
		private Map setters;

		/**
		 * Map of (String) attribute name -> (Method) getter method
		 */
		private Map getters;

		/**
		 * Create a new ClassInfo object for the specified class
		 *
		 * @param clazz
		 */
		public ClassInfo(Class clazz) {

			this.clazz = clazz;
			this.setters = new HashMap();
			this.getters = new HashMap();

			// Cache the getter and setter method info
			Method[] methods = clazz.getMethods();
			String methodName = null;
			Method method = null;
			for (int i = 0; i < methods.length; i++) {
				method = methods[i];
				methodName = method.getName();
				// If it's a getter or setter, cache it.
				if (methodName.startsWith("get")) {
					if ((method.getParameterTypes().length==0) &&
							(!method.getReturnType().equals(Void.class)) &&
							(!methodName.equals("getClass"))) {
						getters.put(methodName.substring(3), method);
					}
				} else if (methodName.startsWith("set")) {
					if (method.getParameterTypes().length==1) {
						if (setters.containsKey(methodName.substring(3))) {
							throw new RuntimeException("Overloaded setters not handled!");
						}
						setters.put(methodName.substring(3), method);
					}
				} else if (methodName.startsWith("is")) {
					if ((method.getParameterTypes().length==0) &&
							(method.getReturnType().equals(Boolean.class) ||
								method.getReturnType().equals(Boolean.TYPE))) {
						getters.put(methodName.substring(2), method);
					}
				}
			}
		}

		/**
		 * Returns all the getter methods.
		 *
		 * @return
		 */
		public Map getGetterMethods() {
			return getters;
		}

		/**
		 * Returns a getter method for the specified attribute if it exists.
		 *
		 * @param attributeName
		 * @return the requested Method or null if none found
		 */
		public Method getGetterMethod(String attributeName) {
			return (Method) getters.get(attributeName);
		}

		/**
		 * Returns a setter method of the specified attribute if it exists.
		 *
		 * @param attribute
		 * @return the requested Method or null if none found
		 */
		public Method getSetterMethod(String attributeName) {
			return (Method) setters.get(attributeName);
		}

		/**
		 * Returns a String representation of the state of this object.
		 *
		 * @return
		 */
		public String toString() {
			StringBuffer buf = new StringBuffer();
			buf.append("[ClassInfo");
			buf.append(" clazz=").append(clazz);
			buf.append(" getters=").append(getters);
			buf.append(" setters=").append(setters);
			return buf.toString();
		}

	}

}
