<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=DAOUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.dao -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.dao;

import java.beans.Introspector;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.security.Principal;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import ${project.rootPackage}.util.ApplicationProperties;

/**
 * This class provides utitlity data access methods.
 */
public class DAOUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(DAOUtils.class);

	/**
	 * Returns a user identifier for the specified user (possibly null).
	 *
	 * @param user - the user
	 * @return
	 */
	public static String getUid(Principal user) {
		if (user!=null) {
			// There is user, use the name
			return user.getName();
		} else {
			// No user, use the guest user id
			return ApplicationProperties.getInstance().
				getStringProperty(ApplicationProperties.DAO_GUESTUSERID, true);
		}
	}

	/**
	 * Creates a list of beans for the specified type and populates it using
	 * data from supplied ResultSet.
	 *
	 * An instance of the specified class is created for each row in the
	 * supplied results. For each column in the result set an attempt is made
	 * to find a corresponding setter on the class. It the setter method
	 * accepts a parameter of type XXX, then the corresponding getXXX() on
	 * the result set will be called to get the column value from the results.
	 *
	 * @param clazz - the class of the bean to extract into.
	 * @param rs - the ResultSet containing the data.
	 * @return a list of populated beans.
	 */
	public static ArrayList extract(Class clazz, ResultSet rs) throws DAOException {

		long startTime = System.currentTimeMillis();
		ArrayList list = new ArrayList();

		try {

			if ((rs!=null)&&(rs.next())) {

				// Get the meta data for the results
				ResultSetMetaData metaData = rs.getMetaData();

				// Get the setter methods that will allow instances of clazz
				// to be populated with data from the results
				Method[] methods = getSetterMethods(clazz, metaData);

				// Go through the results and extract the data into a new
				// instance of clazz for each row.
				do {

					// Extract a populate bean from this row
					Object bean = extractBean(clazz, rs, methods);

					// Add the populated bean to the list
					list.add(bean);

				} while (rs.next());

			}

			return list;

		} catch (SQLException e) {

			// Wrap in a DAOException
			logger.error("SQLException during object extraction", e);
			throw new DAOException(DAOException.ERROR_EXTRACTING_TRANSFER_OBJECTS, null, e);

		} finally {

			if (logger.isDebugEnabled()) {
				logger.debug("extracted " + list.size() +
					" records in " + (System.currentTimeMillis()-startTime) + "ms");
			}

		}

	}

	/**
	 * <p>An extension of the extract(Class, ResultSet) method that provides the
	 * ability to return a subset of the results. Results are split into pages
	 * where each page contains pageSize number of records (except the last
	 * page which may contain less). The results for the requested page is
	 * returned as a PageList.
	 *
	 * @param clazz - the class of the bean to extract into.
	 * @param rs - the ResultSet containing the data.
	 * @param pageNumber - the page number where the first page is page 1.
	 * @param pageSize - the maximumn number of records per page.
	 * @return a PageList containing the populated beans.
	 * @exception DAOException - if the is a problem extracting results.
	 */
	public static PagedData extract(Class clazz, ResultSet rs, int pageNumber, int pageSize) throws DAOException {

		long startTime = System.currentTimeMillis();
		if (logger.isDebugEnabled()) {
			logger.debug("extracting PagedData with pageNumber=" + pageNumber + ", and pageSize=" + pageSize);
		}

		// Create the empty PagedData
		PagedData pagedData = new PagedData(pageNumber, pageSize);

		try {

			// Scroll to the very start of the requested page
			int numNextCalls = (pageNumber-1)*pageSize+1;
			for (int i = 0; i < numNextCalls; i++) {
				if ((rs!=null)&&(rs.next())) {
					// Keep scrolling
				} else {
					// No records for the specified page.
					if (pageNumber == 1) {
						// Page 1 is the only page allowed to have no records.
						pagedData.setTotalRecords(0);
						return pagedData;
					} else {
						// Determine max page number and throw page does not exist exception.
						Integer maxPageNumber = new Integer((i-1)/pageSize + 1);
						throw new DAOException(DAOException.PAGE_DOES_NOT_EXIST, maxPageNumber, null);
					}
				}
			}

			// Get the meta data for the results
			ResultSetMetaData metaData = rs.getMetaData();

			// Get the setter methods that will allow instances of clazz
			// to be populated with data from the results
			Method[] methods = getSetterMethods(clazz, metaData);

			// Extract results for the requested page
			boolean hasNext = true;
			for (int i = 0; i < pageSize; i++) {

				// Extract a populate bean from this row
				Object bean = extractBean(clazz, rs, methods);

				// Add the populated bean to the list
				pagedData.add(bean);

				// Move the result set along
				hasNext = rs.next();
				if (!hasNext) {
					// No more results, break out.
					break;
				}

			}

			// Count the rest of the records
			int numLeft = 0;
			while (hasNext) {
				numLeft++;
				hasNext = rs.next();
			}

			// Set the total number of records
			pagedData.setTotalRecords((pageNumber-1)*pageSize + pagedData.size() + numLeft);

			return pagedData;

		} catch (SQLException e) {

			// Wrap in a DAOException
			logger.error("SQLException during object extraction", e);
			throw new DAOException(DAOException.ERROR_EXTRACTING_TRANSFER_OBJECTS, null, e);

		} finally {

			if (logger.isDebugEnabled()) {
				logger.debug("extracted PagedData=" + pagedData +
					" in " + (System.currentTimeMillis()-startTime) + "ms");
			}

		}

	}

	/**
	 * Creates an instance of the specified class and populates it using
	 * data from the current row of the specified ResultSet.
	 *
	 * @param clazz - the type of the object to create and populate
	 * @param rs - containing the data to populate with
	 * @param methods - the setter methods used to set the data
	 * @return
	 * @throws SQLException
	 * @throws DAOException
	 */
	private static Object extractBean(Class clazz,
									  ResultSet rs,
									  Method[] methods)
									  throws DAOException {

		// Create a new bean instance of the specifed class
		Object bean = null;
		try {
			bean = clazz.newInstance();
		} catch (Exception e) {
			logger.error("exception creating instance of clazz=" + clazz, e);
			throw new DAOException(DAOException.ERROR_CREATING_TRANSFER_OBJECT, null, e);
		}

		// Populate it using the setter methods
		Class parameterType = null;
		Method method = null;
		Object value = null;
		for (int i = 0; i < methods.length; i++) {

			// Get the method corresponding to this column (may be null)
			method = methods[i];
			if (method==null) {
				// No method for this column, skip it.
				continue;
			}

			// Get the type for the method
			parameterType = methods[i].getParameterTypes()[0];

			try {

				// Extract a value for this type
				value = extractValueForType(rs, parameterType, i+1);

			} catch (SQLException e) {
				// Wrap in a DAOException
				logger.error("SQLException extracting a value", e);
				logger.error("parameterType=" + parameterType);
				logger.error("columnIndex=" + (i+1));
				throw new DAOException(DAOException.ERROR_EXTRACTING_VALUE, null, e);

			}

			// Invoke the method on the bean with the extracted value
			try {
				method.invoke(bean, new Object[] {value});
			} catch (Exception e) {
				logger.error("exception populating transfer object=" + bean, e);
				throw new DAOException(DAOException.ERROR_POPULATING_TRANSFER_OBJECT, null, e);
			}

		}
		return bean;
	}

	/**
	 * Utility method for closing a connection that may be null
	 * and suppressing any exceptions raised by the close.
	 *
	 * @param conn
	 */
	public static void close(Connection conn) {
		close(null, null, conn);
	}

	/**
	 * Utility method for closing a connection or statement that may
	 * be null and suppressing any exceptions raised by the close.
	 *
	 * @param statement
	 * @param conn
	 */
	public static void close(Statement statement, Connection conn) {
		close(null, statement, conn);
	}

	/**
	 * Utility method for closing a connection, statement or result set
	 * that may be null and suppressing any exceptions raised by the close.
	 *
	 * @param rs
	 * @param statement
	 * @param conn
	 */
	public static void close(ResultSet rs, Statement statement, Connection conn) {
		// Close the results set
		if (rs!=null) {
			try {
				rs.close();
			} catch (Exception e) {
				// do nothing
			}
		}
		// Close the statement
		if (statement!=null) {
			try {
				statement.close();
			} catch (Exception e) {
				// do nothing
			}
		}
		// Close the connection
		if (conn!=null) {
			try {
				conn.close();
			} catch (Exception e) {
				// do nothing
			}
		}
	}

	/**
	 * Returns a new Map populated with (String) property name -> (Object)
	 * property value for all properties of the specified object that are
	 * exposed via standard JavaBean getter methods.
	 *
	 * @param object
	 */
	public static Map getPropertyValues(Object object) throws DAOException {

		// Create the map
		Map values = new HashMap();

		// Get the methods for this class
		Method[] methods = object.getClass().getMethods();
		for (int i = 0; i < methods.length; i++) {
			Method method = methods[i];
			// Make sure it has no parameters
			Class[] parameterTypes = method.getParameterTypes();
			if ((parameterTypes==null) || (parameterTypes.length!=0)) {
				// Only interested in zero arg methods
				continue;
			}
			// Make sure it is a JavaBean getter method
			Class returnType = method.getReturnType();
			String methodName = method.getName();
			String propertyName = null;
			if ((methodName.startsWith("get")) && (methodName.length()>3)) {
				// Regular getter
				propertyName = Introspector.decapitalize(methodName.substring(3));
			} else if ((methodName.startsWith("is")) &&
					((returnType.equals(Boolean.class))||
						(returnType.equals(boolean.class)))) {
				// Is getter
				propertyName = Introspector.decapitalize(methodName.substring(2));
			} else {
				// Not a JavaBean getter method
				continue;
			}
			// Get the property value
			Object propertyValue = null;
			try {
				propertyValue = method.invoke(object, new Object[]{});
			} catch (Exception e) {
				// Wrap in a DAOException
				logger.error("Exception extracting a value", e);
				logger.error("object=" + object);
				logger.error("method=" + method);
				throw new DAOException(DAOException.ERROR_EXTRACTING_VALUE, null, e);
			}
			// Now add it to the map (null values ok)
			values.put(propertyName, propertyValue);
			if (logger.isDebugEnabled()) {
				logger.debug("putting " + propertyName + "=" + propertyValue);
				if (propertyValue!=null) {
					logger.debug(" value class=" + propertyValue.getClass().getName());
				}
			}
		}

		// Return the populated map
		return values;

	}

	// --------------- Start Private Methods ------------------------------ //

	/**
	 * Returns an array of methods for the specified class that can be used
	 * to set data on instances of the class using data from the ResultSet
	 * that the ResultSetMetaData is for.
	 *
	 * @param clazz - the class whose methods we are after
	 * @param metaData - from the ResultSet
	 * @return an array of Methods, one for each column in the meta data,
	 * containing null values for those columns where there was no appropriate
	 * setter method.
	 * @throws SQLException
	 */
	private static Method[] getSetterMethods(Class clazz, ResultSetMetaData metaData) throws SQLException {

		Method[] allMethods = clazz.getMethods();
		Method[] methods = new Method[metaData.getColumnCount()];	// One method for each column
		String columnName = null;
		String methodName = null;
		Method setterMethod = null;

		// Go through the list of column names in the result set and try find
		// a method that can be used to set the values. Note: The first setter
		// with the correct name and accepts a single parameter is chosen. The
		// correct method name is chosen using equalsIgnoreCase because it is
		// possible that, for example, the datbase name "url" is set by "setURL".
		for (int i = 0; i < metaData.getColumnCount(); i++) {
			columnName = metaData.getColumnName(i+1);
			methodName = getJavaSetterNameFromColumnName(columnName);
			boolean foundSetter = false;
			for (int j = 0; j < allMethods.length; j++) {
				setterMethod = allMethods[j];
				if ((setterMethod.getName().equalsIgnoreCase(methodName)) &&
					(setterMethod.getParameterTypes().length==1)) {
					// Found a setter with correct name and one param
					methods[i] = setterMethod;
					foundSetter = true;
					break;
				}
			}
			// Debug
			if (logger.isDebugEnabled()) {
				if (!foundSetter) {
					logger.debug("did not find setter method for columnName=" + columnName);
				} else {
					logger.debug("found setter=" + setterMethod +" for columnName=" + columnName);
				}
			}
		}

		return methods;
	}

	/**
	 * CUSTOMER_ID -> setCustomerId
	 */
	private static String getJavaSetterNameFromColumnName(String columnName) {
		return ("set"+getJavaStyleFromDatabaseStyle(columnName, true));
	}

	/**
	 * Returns a java style name from the specified database styled string.
	 *
	 * customer_id -> CustomerId
	 * or
	 * customer_id -> customerId
	 */
	private static String getJavaStyleFromDatabaseStyle(String name, boolean uppercaseFirst) {
		StringBuffer buf = new StringBuffer(name.length());
		int length = name.length();
		char c = name.charAt(0);
		buf.append((uppercaseFirst)?(Character.toUpperCase(c)):(Character.toLowerCase(c)));
		boolean uppercaseNext = false;
		for (int i = 1; i < length; i++) {
			c = name.charAt(i);
			if (c == '_') {
				// Uppercase the next non '_' char
				uppercaseNext = true;
			} else if (uppercaseNext) {
				buf.append(Character.toUpperCase(c));
				uppercaseNext = false;
			} else {
				buf.append(Character.toLowerCase(c));
			}
		}
		return buf.toString();
	}

	/**
	 * Extract a value of the specified type from the supplied result.
	 *
	 * @param rs - the ResultSet containing the data.
	 * @param type - the type of the value that is required.
	 * @param columnIndex - the index of the column in the ResultSet.
	 * @return the value that was extracted.
	 * @throws SQLException - if there was a problem getting the value.
	 */
	private static Object extractValueForType(ResultSet rs,
											  Class type,
											  int columnIndex)
											  throws SQLException {

		// Extract values where there is a getXXX() method returning an Object
		if (type.equals(BigDecimal.class)) {
			return rs.getBigDecimal(columnIndex);
		} else if (type.equals(String.class)) {
			return rs.getString(columnIndex);
		} else if (type.equals(Date.class)) {
			return rs.getDate(columnIndex);
		} else if (type.equals(Time.class)) {
			return rs.getTime(columnIndex);
		} else if (type.equals(Timestamp.class)) {
			return rs.getTimestamp(columnIndex);
		} else if (type.equals(byte[].class)) {
			return rs.getBytes(columnIndex);
		} else if (type.equals(Blob.class)) {
			return rs.getBlob(columnIndex);
		} else if (type.equals(Clob.class)) {
			return rs.getClob(columnIndex);
		}

		// Where there is no getXXX() for the specified type that returns an
		// Object we must first call getObject() to see if the value is null.
		// NOTE: Depending on the driver implementation this may require twice
		// as much work as simply calling the getXXX() method, but this way we
		// can correctly and easily handle nulls.

		// See if the value is null
		if (rs.getObject(columnIndex)==null) {
			return null;
		}

		// The value is not null. Get it in the requested type
		if (type.equals(Byte.class)) {
			return new Byte(rs.getByte(columnIndex));
		} else if (type.equals(Short.class)) {
			return new Short(rs.getShort(columnIndex));
		} else if (type.equals(Integer.class)) {
			return new Integer(rs.getInt(columnIndex));
		} else if (type.equals(Long.class)) {
			return new Long(rs.getLong(columnIndex));
		} else if (type.equals(Double.class)) {
			return new Double(rs.getDouble(columnIndex));
		} else if (type.equals(Float.class)) {
			return new Float(rs.getFloat(columnIndex));
		} else if (type.equals(Boolean.class)) {
			return new Boolean(rs.getBoolean(columnIndex));
		} else {
			throw new IllegalArgumentException("Unhandled type '" + type + "'");
		}

	}

	// --------------- End Private Methods -------------------------------- //

}
