<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=DatabaseUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.database -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.database;

import java.sql.Connection;
import java.sql.SQLException;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import javax.transaction.UserTransaction;

import org.apache.log4j.Logger;

import ${project.rootPackage}.util.ApplicationProperties;

/**
 * This class provides easy access to Connections from a configurable DataSource
 * and simplified UserTransaction management.
 */
public class DatabaseUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(DatabaseUtils.class);

	/**
	 * Reference to the lazy loaded DataSource
	 */
	private static DataSource dataSource = null;

	/**
	 * Calls begin() on the current UserTransaction
	 *
	 * @exception DatabaseException if the UserTransaction cannot be located
	 * or there is an error calling begin().
	 */
	public static void begin() throws DatabaseException {

		logger.debug("beginning transaction..");

		// Get the current UserTransaction
		UserTransaction ut = getUserTransaction();

		try {

			// Begin the transaction
			ut.begin();
			logger.debug("transaction has commenced");

		} catch (Exception e) {

			// Wrap the exception up in a DatabaseException
			logger.error("exception beginning user transaction", e);
			throw new DatabaseException(
				DatabaseException.ERROR_BEGINNING_USER_TRANSACTION, null, e);

		}

	}

	/**
	 * Calls commit() on the current UserTransaction
	 *
	 * @exception DatabaseException if the UserTransaction cannot be located
	 * or there is an error calling commit().
	 */
	public static void commit() throws DatabaseException {

		logger.debug("commiting transaction..");

		// Get the current UserTransaction
		UserTransaction ut = getUserTransaction();

		try {

			// Commit the transaction
			ut.commit();
			logger.debug("transaction committed");

		} catch (Exception e) {

			// Try and roll it back (ignore errors)
			try {
				ut.rollback();
			} catch (Exception e2) {}

			// Wrap the exception up in a DatabaseException
			logger.error("exception commiting transaction", e);
			throw new DatabaseException(
				DatabaseException.ERROR_COMMITTING_USER_TRANSACTION, null, e);

		}

	}


	/**
	 * Calls rollback() on the current UserTransaction.
	 *
	 * @exception DatabaseException if the UserTransaction cannot be located
	 * or there is an error calling rollback().
	 */
	public static void rollback() throws DatabaseException {

		logger.debug("rolling back...");

		// Get the current UserTransaction
		UserTransaction ut = getUserTransaction();

		try {

			// And try roll it back
			ut.rollback();
			logger.debug("rolled back!");

		} catch (Exception e) {

			// Wrap the exception up in a DatabaseException
			logger.error("exception rolling back transaction", e);
			throw new DatabaseException(
				DatabaseException.ERROR_LOCATING_USER_TRANSACTION, null, e);

		}

	}

	/**
	 * Gets a connection from the DataSource.
	 *
	 * @return a Connection
	 * @throws SQLException if there was a problem getting the Connection
	 * @throws DatabaseException if there is a problem locating the DataSource
	 */
	public static Connection getConnection() throws SQLException, DatabaseException {
		Connection conn = getDataSource().getConnection();
		if (logger.isDebugEnabled()) logger.debug("returning conn=" + conn);
		return conn;
	}

	/**
	 * Returns a lazy located DataSource configured from application properties.
	 *
	 * @return the DataSource
	 * @throws DatabaseException if there is a problem locating the DataSource.
	 */
	private static DataSource getDataSource() throws DatabaseException {

		if (dataSource==null) {

			// Get the jndi name under which the data source is located
			String jndiName = ApplicationProperties.getInstance().
				getStringProperty(ApplicationProperties.DATASOURCE_JNDINAME, true);

			try {

				// Create a default initial context
				InitialContext ctx = new InitialContext();

				// Lookup the DataSource
				dataSource =(DataSource)ctx.lookup("java:comp/env/" + jndiName);

				// Throw an Exception if not found
				if (dataSource==null) {
					throw new DatabaseException(
							DatabaseException.COULD_NOT_FIND_DATA_SOURCE,
							jndiName, null);
				}

				logger.debug("initialised dataSource=" + dataSource);

			} catch (NamingException ne) {

				// Wrap the exception up in a DatabaseException
				logger.error("exception looking up datasource " + jndiName, ne);
				throw new DatabaseException(
					DatabaseException.ERROR_LOCATING_DATA_SOURCE, null, ne);

			}

		}

		return dataSource;

	}

	/**
	 * Retrieves the current UserTransaction from jndi.
	 *
	 * @return the current UserTransaction
	 * @throws NamingException
	 */
	private static UserTransaction getUserTransaction() throws DatabaseException {

		UserTransaction ut = null;
		InitialContext ctx = null;
		try {

			// Create a default initial context
			ctx = new InitialContext();

			// Try locate the UserTransaction
			ut = (UserTransaction) ctx.lookup("java:comp/UserTransaction");
			if (logger.isDebugEnabled()) {
				logger.debug("located UserTransaction=" + ut);
			}

		} catch (NamingException ne) {

			// Wrap the exception up in a DatabaseException
			logger.error("naming exception looking up user transaction", ne);
			throw new DatabaseException(
				DatabaseException.ERROR_LOCATING_USER_TRANSACTION, null, ne);

		}

		return ut;
	}

}
