<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=ExtendedProperties.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util;

import java.io.IOException;
import java.io.InputStream;
import java.util.MissingResourceException;
import java.util.Properties;

import org.apache.log4j.Logger;

/**
 * An extension of the standard Properties that provides getXXXProperty()
 * methods for common Java types and convenience method for loading properties
 * from a resource.
 */
public class ExtendedProperties extends Properties {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(ExtendedProperties.class);

	/**
	 * Contructs a new empty ExtendedProperties object.
	 */
	public ExtendedProperties() {
		super();
	}

	/**
	 * Constructs a new empty ExtendedProperties with the specified default values
	 *
	 * @param defaults
	 */
	public ExtendedProperties(Properties defaults) {
		super(defaults);
	}

	/**
	 * Returns a String for the specified property.
	 *
	 * @param name - the name of the property
	 * @param required - whether to throw exception if value not found
	 * @exception MissingResourceException if value not found and was required
	 * @return the value of the property
	 */
	public String getStringProperty(String name, boolean required) throws MissingResourceException {
		String value = getProperty(name);
		if (required && value==null) {
			throw new MissingResourceException("Missing property '" + name + "'", "", "");
		}
		return value;
	}

	/**
	 * Returns an Integer for the specified property.
	 *
	 * @param name - the name of the property
	 * @param required - whether to throw exception if value not found
	 * @exception MissingResourceException if value not found and was required
	 * @return the value of the property as an Integer
	 */
	public Integer getIntegerProperty(String name, boolean required) throws MissingResourceException {
		String value = getProperty(name);
		if (value!=null) {
			try {
				return new Integer(value);
			} catch (NumberFormatException nfe) {
				if (logger.isDebugEnabled()) {
					logger.warn("could not parse integer for property '" + name + "'");
				}
			}
		}
		if (required && value==null) {
			throw new MissingResourceException("Missing property '" + name + "'", "", "");
		}
		return null;
	}

	/**
	 * Delegates to getIntegerProperty(name) and if this returns null then
	 * the specified default value is returned.
	 *
	 * @param name - the name of the property
	 * @param defaultValue - the value to return if none found
	 * @return the value of the property as an Integer
	 */
	public Integer getIntegerProperty(String name, int defaultValue) {
		Integer value = getIntegerProperty(name, false);
		return ((value==null)?(new Integer(defaultValue)):(value));
	}

	/**
	 * Returns a Long for the specified property. If the property value
	 * is not present or it cannot be then null is returned.
	 *
	 * @param name - the name of the property
	 * @param required - whether to throw exception if value not found
	 * @exception MissingResourceException if value not found and was required
	 * @return the value of the property as a Long
	 */
	public Long getLongProperty(String name, boolean required) throws MissingResourceException {
		String value = getProperty(name);
		if (value!=null) {
			try {
				return new Long(value);
			} catch (NumberFormatException nfe) {
				if (logger.isDebugEnabled()) {
					logger.warn("could not parse long for property '" + name + "'");
				}
			}
		}
		if (required && value==null) {
			throw new MissingResourceException("Missing property '" + name + "'", "", "");
		}
		return null;
	}

	/**
	 * Delegates to getLongProperty(name) and if this returns null then
	 * the specified default value is returned.
	 *
	 * @param name - the name of the property
	 * @param defaultValue - the value to return if none found
	 * @return the value of the property as a Long
	 */
	public Long getLongProperty(String name, long defaultValue) {
		Long value = getLongProperty(name, false);
		return ((value==null)?(new Long(defaultValue)):(value));
	}

	/**
	 * Returns a Float for the specified property. If the property value
	 * is not present or it cannot be then null is returned.
	 *
	 * @param name - the name of the property
	 * @param required - whether to throw exception if value not found
	 * @exception MissingResourceException if value not found and was required
	 * @return the value of the property as a Float
	 */
	public Float getFloatProperty(String name, boolean required) throws MissingResourceException {
		String value = getProperty(name);
		if (value!=null) {
			try {
				return new Float(value);
			} catch (NumberFormatException nfe) {
				if (logger.isDebugEnabled()) {
					logger.warn("could not parse float for property '" + name + "'");
				}
			}
		}
		return null;
	}

	/**
	 * Delegates to getFloatProperty(name) and if this returns null then
	 * the specified default value is returned.
	 *
	 * @param name - the name of the property
	 * @param defaultValue - the value to return if none found
	 * @return the value of the property as a Float
	 */
	public Float getFloatProperty(String name, float defaultValue) {
		Float value = getFloatProperty(name, false);
		return ((value==null)?(new Float(defaultValue)):(value));
	}

	/**
	 * Returns a Double for the specified property. If the property value
	 * is not present or it cannot be then null is returned.
	 *
	 * @param name - the name of the property
	 * @param required - whether to throw exception if value not found
	 * @exception MissingResourceException if value not found and was required
	 * @return the value of the property as a Double
	 */
	public Double getDoubleProperty(String name, boolean required) throws MissingResourceException {
		String value = getProperty(name);
		if (value!=null) {
			try {
				return new Double(value);
			} catch (NumberFormatException nfe) {
				if (logger.isDebugEnabled()) {
					logger.warn("could not parse double for property '" + name + "'");
				}
			}
		}
		if (required && value==null) {
			throw new MissingResourceException("Missing property '" + name + "'", "", "");
		}
		return null;
	}

	/**
	 * Delegates to getDoubleProperty(name) and if this returns null then
	 * the specified default value is returned.
	 *
	 * @param name - the name of the property
	 * @param defaultValue - the value to return if none found
	 * @return the value of the property as a Double
	 */
	public Double getDoubleProperty(String name, double defaultValue) {
		Double value = getDoubleProperty(name, false);
		return ((value==null)?(new Double(defaultValue)):(value));
	}

	/**
	 * <p>Returns a new Properties object populated using the specified resource.
	 * The resource is loaded using the supplied class's getResourceAsStream()
	 * method.
	 *
	 * <p>Class.getResourceAsStream() requires the name to seperate packages with
	 * "/". If the name starts with "/" then it is treated as absolute, else it
	 * is treated as relative to the class's package.
	 *
	 * <p>For example, "/com/company/project/application.properties" specifies an
	 * absoulte resource and "application.properties" is relative to class's
	 * package.
	 *
	 * @param name - the resource name
	 * @param clazz - used to get the resource using getResourceAsStream()
	 */
	public void load(String name, Class clazz) {

		InputStream is = null;

		try {

			// Get the resource as an InputStream
			is = clazz.getResourceAsStream(name);

			// Populate this using the InputStream
			if (is != null) {
				// The resource was found
				try {
					load(is);
				} catch (IOException ioe) {
					// This will *very* rarely happen. Rather than throw the
					// exception up as a checked exception and force everyone
					// who uses it to handle the exception, log it and throw
					// up a RuntimeException.
					logger.error("exception loading properties from resource '" + name + "'", ioe);
					throw new RuntimeException("IOException loading properties: " + ioe.getMessage());
				}

			} else {
				// The resource is missing
				logger.warn("could not locate resource '" + name + "' using class '" + clazz.getName() + "'");
			}

		} finally {

			if (is != null) {
				try {
					is.close();
				} catch (Exception e) {
					// Ignore
				}
			}
		}

	}

}
