<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=SequenceGenerator.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.database -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.database;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import ${project.rootPackage}.util.ApplicationProperties;

/**
 * <p>The SequenceGenerator is a utility class that can be used to generate
 * unique integer values for arbitrary names.
 *
 * <p>It utilises a cache to reduce the number of database hits required.
 * Because of the cache there is no guarantee that all numbers will be
 * used. For example, if the program is shut down while there are some
 * unused numbers in the cache, they will be lost. One way of preventing
 * this is to use a cache size of 1.
 */
public class SequenceGenerator {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(SequenceGenerator.class);

	/**
	 * The table used to store sequence information.
	 */
	public static final String TABLE_NAME = "Sequence";

	/**
	 * The field used to store the last id for a name.
	 */
	public static final String LAST_ID_FIELD = "last_id";

	/**
	 * The field used to store the name.
	 */
	public static final String NAME_FIELD = "name";

	/**
	 * The sql used to select the next id for a name.
	 */
	public static final String SELECT_SQL = "SELECT " + LAST_ID_FIELD +
			" FROM " + TABLE_NAME + " WHERE " + NAME_FIELD + " = ?";

	/**
	 * The sql used to update the next id for a name.
	 */
	public static final String UPDATE_SQL = "UPDATE " + TABLE_NAME + " SET " +
			LAST_ID_FIELD + " = ? WHERE " + NAME_FIELD + " = ? AND " + LAST_ID_FIELD + " = ?";

	/**
	 * The sql used to insert the initial id for a name.
	 */
	public static final String INSERT_SQL = "INSERT INTO " + TABLE_NAME +
			"(" + NAME_FIELD + ", " + LAST_ID_FIELD + ") VALUES (?,?)";

	/**
	 * The number of IDs to grab from the database when there
	 * are none left to allocate.
	 */
	private static int grabSize;

	/**
	 * Map of (String) name -> (Long) lastGrabbedId
	 */
	private static Map lastGrabbedIds = new HashMap();

	/**
	 * Map of (String) name -> (Long) nextId
	 */
	private static Map nextIds = new HashMap();

	static {

		//Initialise the grab size from the application properties
		grabSize = ApplicationProperties.getInstance().getIntegerProperty(
			ApplicationProperties.SEQUENCEGENERATOR_GRABSIZE, 10).intValue();

		if (logger.isDebugEnabled()) {
			logger.debug("determined grabSize=" + grabSize);
		}

	}

	/**
	 * Returns next id for the specified name.
	 *
	 * @param name - the name to get the next id for
	 * @exception SQLException - if there is a problem grabbing IDs
	 * @exception DatabaseException - if there is a problem getting a Connection
	 * @return
	 */
	public static synchronized Long getNextId(String name) throws DatabaseException, SQLException {

		// See if there is a cached id for the specified name
		if (!nextIds.containsKey(name)) {
			// No cached id. Init the cached for the specified name
			grabMoreIds(name);
		}

		// Get the nextId and lastGrabbedId for the specified name
		long lastGrabbedId = ((Long) lastGrabbedIds.get(name)).longValue();
		long nextId = ((Long) nextIds.get(name)).longValue();
		if (nextId>lastGrabbedId) {
			// We have run out of cached IDs. Grab some more
			grabMoreIds(name);
			// Get the next id
			nextId = ((Long) nextIds.get(name)).longValue();
		}

		// Update the nextId for the specified name
		nextIds.put(name, new Long(nextId+1));

		if (logger.isDebugEnabled()) {
			logger.debug("returning nextId=" + nextId + " for name=" + name);
		}

		// Return the next id
		return new Long(nextId);

	}

	/**
	 * Caches some more IDs for the specified name.
	 *
	 * @param name - the name to grab more IDs for
	 * @exception DatabaseException if there is a problem getting a Connection
	 * @exception SQLException if there is a problem executing the sql
	 */
	private static void grabMoreIds(String name) throws DatabaseException, SQLException {

		// Update the record in the sequence table using specfied name
		PreparedStatement ps = null;
		Connection conn = null;
		ResultSet rs = null;
		int updateCount = 0;
		long lastId = 0;
		int i = 0;

		if (logger.isDebugEnabled()) logger.debug("grabbing IDs for name=" + name);

		try {

			// Get a connection
			conn = DatabaseUtils.getConnection();

			do {

				// Select the last id for the name (may be none)
				ps = conn.prepareStatement(SELECT_SQL);
				ps.setString(1,name);
				rs = ps.executeQuery();
				if (rs.next()) {

					// There is a last_id in the database for name, get it
					lastId = rs.getLong(LAST_ID_FIELD);
					rs.close();
					ps.close();

					// Update with the new value. NOTE: use a where clause in
					// the update to ensure that we have the absolute most
					// recent last_id.
					if (logger.isDebugEnabled()) logger.debug("updating existing sequence record");
					ps = conn.prepareStatement(UPDATE_SQL);
					ps.setLong(1,lastId+grabSize);
					ps.setString(2,name);
					ps.setLong(3,lastId);
					updateCount = ps.executeUpdate();
					ps.close();

				} else {

					// No id in the database for the specified name
					lastId = 0;
					rs.close();
					ps.close();

					// Insert the initial values. NOTE: It is important that
					// the Sequence table has a unique contrainst on the name
					// column to prevent duplicate initial values from being
					// inserted.
					if (logger.isDebugEnabled()) logger.debug("inserting new sequence record");
					ps = conn.prepareStatement(INSERT_SQL);
					ps.setString(1,name);
					ps.setInt(2,grabSize);
					updateCount = ps.executeUpdate();
					ps.close();

				}

				// Repeat until we get a successful update. NOTE: limit to 100
				// because if the update has not worked by then there is
				// probably else something wrong.

			} while ((updateCount != 1) && (i++ < 100));

			// If the update was not successful then throw an error
			if (updateCount != 1) {
				if (logger.isDebugEnabled()) logger.error("could not update sequence table!");
				lastGrabbedIds.put(name,new Long(-1));
				nextIds.put(name, new Long(-1));
			} else {
				// Successfully grabbed some IDs. Update the caches
				if (logger.isDebugEnabled()) logger.debug("grabbed " + grabSize + " IDs starting from " + (lastId+1));
				lastGrabbedIds.put(name,new Long(lastId+grabSize));
				nextIds.put(name, new Long(lastId+1));
			}

		} finally {
			if (conn!=null) {
				try {
					conn.close();
				} catch (Exception e) {}
			}
		}

	}

}
