<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=SqlUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.database -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.database;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * Provides some SQL related functions.
 */
public class SqlUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(SqlUtils.class);

	/**
	 * The character that prefixes named parameters in the sql. For example:<br>
	 *
	 * Select * from CUSTOMER where CUSTOMER_ID = :customerId
	 */
	public static final char PARAMETER_PREFIX = ':';

	/**
	 * <p>Converts a sql string containing named parameters the format required
	 * by prepared statements and populates the valuesToBind collection with
	 * values from the parameterValues map in the correct order that they need to be
	 * bound in resulting sql.
	 *
	 * @param sql - the sql containing named parameters
`	 * @param parameterValues - a map of (String) name -> (object) value (possibly null)
	 * @param valuesToBind - empty list that will be populated with values
	 * @return the sql in prepared statement format
	 */
	public static String constructPreparedStatementSQL(String sql, Map parameterValues, List valuesToBind) {

		StringBuffer buf = new StringBuffer(sql.length());
		Set usedParameterNames = new HashSet();
		String parameterName = null;
		Object parameterValue = null;
		int fromIndex = 0;
		int prefixIndex = sql.indexOf(PARAMETER_PREFIX);

		while (prefixIndex!=-1) {

			// There is possibly an unprocessed parameter starting at prefixIndex

			// Add the sql from fromIndex to prefixIndex to the buffer as this is
			// just raw sql with no parameter replacements needed.
			buf.append(sql.substring(fromIndex,prefixIndex));

			// Check if there are any characters after the located PARAMETER_PREFIX
			if (prefixIndex+1 > sql.length()) {
				// There are no more chars after the PARAMETER_PREFIX, so
				// its not a valid parameter name. Just add the prefix to the buffer.
				buf.append(PARAMETER_PREFIX);

			} else {
				// There are characters after the PARAMETER_PREFIX, examine them
				// in order to determine the parameter name.
				int length = 1;
				if (Character.isJavaIdentifierStart(sql.charAt(prefixIndex+length))) {

					// First char is valid, keep going till we find an invalid one
					// or we reach the end of the sql string.
					do {
						length++;
					} while ((prefixIndex+length<sql.length())
						&& (Character.isJavaIdentifierPart(sql.charAt(prefixIndex+length))));

					// We have found the end of the parameter name
					parameterName = sql.substring(prefixIndex+1, prefixIndex+length);
					if ((parameterValues != null) && (parameterValues.containsKey(parameterName))) {

						// There is a value in the parameterValues map. Add a ? to the sql in place
						// of the parameter name and add the value to the valuesToBind collection.
						parameterValue = parameterValues.get(parameterName);
						logger.debug("found " + parameterName + "=" + parameterValue);

						// Records that we have used this parameter name
						usedParameterNames.add(parameterName);

						// Append a '?' where the parameter value will be bound
						if ((parameterValue!=null)&&(parameterValue instanceof Collection)) {
							// Special case for collections, create a set, ie (?,?,?)
							buf.append("(");
							for (Iterator iter = ((Collection)parameterValue).iterator(); iter.hasNext(); ) {
								valuesToBind.add(iter.next());
								buf.append("?");
								if (iter.hasNext()) buf.append(",");
							}
							buf.append(")");
						} else {
							// Treat it as normally expected
							buf.append("?");
							valuesToBind.add(parameterValue);
						}

					} else {

						// There is no value for this parameter. It is possible that the
						// user was just using the : in the sql, so let it go through.
						// It will most like give an error later on if it was not correct.
						logger.warn("values map contains no value for parameterName=" + parameterName);
						buf.append(PARAMETER_PREFIX).append(parameterName);

					}

					// We have processed the parameter name. Set the fromIndex to point
					// to the next character after the parameter name.
					fromIndex = prefixIndex + length;

				} else {
					// Not a valid parameter name. Just ignore it.
					// Add the PARAMETER_PREFIX to the buffer and move along.
					buf.append(PARAMETER_PREFIX);
					// Set the fromIndex to be the next character after the PARAMETER_PREFIX
					fromIndex = prefixIndex+1;

				}

			}

			// See if there are any more PARAMETER_PREFIX characters..
			prefixIndex = sql.indexOf(PARAMETER_PREFIX, fromIndex);

		}

		// Check to make sure we used all the values
		if (logger.isDebugEnabled()) {
			if (parameterValues!=null) {
				Set unusedParameterNames = new HashSet(parameterValues.keySet());
				unusedParameterNames.removeAll(usedParameterNames);
				if (!unusedParameterNames.isEmpty()) {
					// There were some unused values
					logger.debug("unused parameter names=" + unusedParameterNames);
				}
			}
		}

		// Append the rest of the sql from fromIndex to complete processing
		buf.append(sql.substring(fromIndex));

		// Debug out the sql and parameters to give the user an idea of what
		// executed. It will not be exactly the same because it just uses
		// a best guess at how the different types will be rendered.
		if (logger.isDebugEnabled()) debugSql(buf.toString(), valuesToBind);

		// So here we are. The buffer now contains the sql in the correct form
		// for creating a prepared statement, and the collection of
		// valuesToBind has the parameters to be bound in the correct order.
		// Return the contents of the buffer.
		return buf.toString();

	}

	/**
	 * <p>This method debugs the approximate sql that will be generated
	 * by a PreparedStatement that uses the specified sql and values.
	 *
	 * <p>NOTE: This is only approximate because it is up to the JDBC
	 * Driver as to the exact conversion.
	 *
	 * @param preparedStatmentSql the sql in prepared statement format
	 * @param valuesToBind the values
	 */
	private static void debugSql(String preparedStatmentSql, List valuesToBind) {

		StringBuffer buf = new StringBuffer(preparedStatmentSql.length());
		int index = 0;
		for (Iterator iter = valuesToBind.iterator(); iter.hasNext();) {
			// Get the next value
			Object value = iter.next();
			// Determine the next index of the '?'
			int nextIndex = preparedStatmentSql.indexOf('?',index);
			if (nextIndex == -1) {
				// There is no next index, meaning there are more values
				// than there are places for them. Just break.
				break;
			}
			// Append the sql up to the index of the next value
			buf.append(preparedStatmentSql.substring(index, nextIndex));
			// Quote the value if needed
			if ((value instanceof String) || (value instanceof java.util.Date) ||
					(value instanceof java.sql.Date) || (value instanceof Timestamp)) {
				// These types need to be surrounded by quotes
				value = ("'"+value+"'");
			}
			// Append the value
			buf.append(value);
			// Move past the '?'
			index = nextIndex+1;
		}
		// Append the rest of the sql after the last index
		buf.append(preparedStatmentSql.substring(index));
		// Finally debug the sql as requested
		logger.debug("sql is approximately=\n" + buf);

	}

}
