<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=StatementPreparer.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.database -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.database;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.digester.Digester;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import ${project.rootPackage}.util.template.TemplateException;
import ${project.rootPackage}.util.template.TemplateUtils;

/**
 * <p>This class is used to construct PreparedStatements in a flexible way.
 *
 * <p>Template sql is loaded from resource files. When a request is made at
 * runtime to create a PreparedStatement, the specified sql template is
 * processed using the specified values to produce produce a parameterised
 * sql statement. This is then converted into prepared statement sql using the
 * values and finally a PreparedStatement is created.
 */
public class StatementPreparer {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(StatementPreparer.class);

	/**
	 * The map of sql for this StatementPreparer.
	 *
	 * It contains (String) sqlKey -> (String) sql template.
	 */
	protected Map sqlMap = new Hashtable();

	/**
	 * Default constructor
	 */
	public StatementPreparer() {}

	/**
	 * Creates a PreparedStatement using the sql specified by sqlKey and bound
	 * with the specified values.
	 *
	 * @param conn - the Connection to create the PreparedStatement with
	 * @param sqlKey - identifying the sql template from the resource
	 * @param parameter1 - the name of the first parameter.
	 * @param value1 - the value of the first parameter.
	 * @return the new PreparedStatement.
	 * @exception SQLException if there is a problem preparing the statement.
	 * @exception StatementPreparerException if there is a problem reading the
	 * sql resources or there is no sql for the specified sql key.
	 * @exception TemplateException if there is a problem processing the template.
	 */
	public PreparedStatement prepare(
			Connection conn,
			String sqlKey,
			String parameter1,
			Object value1)
			throws TemplateException,
				StatementPreparerException,
				SQLException {

		Map parameters = new HashMap();
		parameters.put(parameter1, value1);
		return prepare(conn, sqlKey, parameters, -1, -1);
	}

	/**
	 * Creates a PreparedStatement bound with the specified values using the
	 * specified sql.
	 *
	 * @param conn - the Connection to create the PreparedStatement with
	 * @param sqlKey - identifying the sql template from the resource
	 * @param parameter1 - the name of the first parameter.
	 * @param value1 - the value of the first parameter.
	 * @param parameter2 - the name of the second parameter.
	 * @param value2 - the value of the second parameter.
	 * @return the new PreparedStatement.
	 * @exception SQLException if there is a problem preparing the statement.
	 * @exception StatementPreparerException if there is a problem reading the
	 * sql resources or there is no sql for the specified sql key.
	 * @exception TemplateException if there is a problem processing the template.
	 */
	public PreparedStatement prepare(
			Connection conn,
			String sqlKey,
			String parameter1,
			Object value1,
			String parameter2,
			Object value2)
			throws TemplateException,
				StatementPreparerException,
				SQLException {

		Map parameters = new HashMap();
		parameters.put(parameter1, value1);
		parameters.put(parameter2, value2);
		return prepare(conn, sqlKey, parameters, -1, -1);
	}

	/**
	 * Result sets created using the returned PreparedStatement will have
	 * forward-only type and read-only concurrency, by default.
	 *
	 * @param conn - the Connection to create the PreparedStatement with
	 * @param sqlKey - identifying the sql template from the resource
	 * @return the new PreparedStatement.
	 * @exception SQLException if there is a problem preparing the statement.
	 * @exception StatementPreparerException if there is a problem reading the
	 * sql resources or there is no sql for the specified sql key.
	 * @exception TemplateException if there is a problem processing the template.
	 */
	public PreparedStatement prepare(
			Connection conn,
			String sqlKey)
			throws TemplateException,
				StatementPreparerException,
				SQLException {

		return prepare(conn, sqlKey, null, -1, -1);
	}

	/**
	 * Result sets created using the returned PreparedStatement will have
	 * forward-only type and read-only concurrency, by default.
	 *
	 * @param conn - the Connection to create the PreparedStatement with
	 * @param sqlKey - identifying the sql template from the resource
	 * @param values - a Map of (String) parameter -> (Object) value
	 * @return the new PreparedStatement
	 * @exception SQLException if there is a problem preparing the statement.
	 * @exception StatementPreparerException if there is a problem reading the
	 * sql resources or there is no sql for the specified sql key.
	 * @exception TemplateException if there is a problem processing the template.
	 */
	public PreparedStatement prepare(
			Connection conn,
			String sqlKey,
			Map values)
			throws TemplateException,
				StatementPreparerException,
				SQLException {

		return prepare(conn, sqlKey, values, -1, -1);
	}

	/**
	 * Creates a PreparedStatement using the specified parameter values and sql.
	 *
	 * @param conn - the Connection to create the PreparedStatement with
	 * @param sqlKey - identifying the sql template from the resource
	 * @param parameterValues - a Map of (String) parameter -> (Object) value
	 * @param resultSetType - a result set type; see ResultSet.TYPE_XXX
	 * @param resultSetConcurrency - a concurrency type; see ResultSet.CONCUR_XXX
	 * @return the new PreparedStatement.
	 * @exception SQLException if there is a problem preparing the statement.
	 * @exception StatementPreparerException if there is a problem reading the
	 * sql resources or there is no sql for the specified sql key.
	 * @exception TemplateException if there is a problem processing the template.
	 */
	public PreparedStatement prepare(
			Connection conn,
			String sqlKey,
			Map parameterValues,
			int resultSetType,
			int resultSetConcurrency )
			throws SQLException,
				TemplateException,
				StatementPreparerException {

		// Debug the parameters
		if (logger.isDebugEnabled()) {
			logger.debug("preparing statement using:");
			logger.debug("conn=" + conn);
			logger.debug("sqlKey=" + sqlKey);
			logger.debug("parameterValues=" + parameterValues);
			logger.debug("resultSetType=" + resultSetType);
			logger.debug("resultSetConcurrency=" + resultSetConcurrency);
		}

		// Get the specified sql template from the cache
		String sql = (String) sqlMap.get(sqlKey);
		if (sql==null) {
			// No sql for this key. Log the error.
			logger.error("could not find sql for sqlKey=" + sqlKey);
			logger.error("sqlMap=" + sqlMap);
			throw new StatementPreparerException(
					StatementPreparerException.MISSING_SQL_KEY, sqlKey, null);
		}

		// Evaluate the sql template using freemarker
		sql = TemplateUtils.process(parameterValues, sql);

		// Now the sql will just contain references to variables in the
		// parameterValues map. For example the object under "customerId" in
		// the values map will be referenced in the sql as :customerId. Need to
		// convert the sql into a form that can be used to directly create the
		// PreparedStatement. Also create the list of parameters to bind to it
		// in the correct order.
		List valuesToBind = new ArrayList();
		sql = SqlUtils.constructPreparedStatementSQL(sql, parameterValues, valuesToBind);

		// Create the PreparedStatement, bind the parameters. Only use result
		// set type and concurrency if they have been specified. Else leave it
		// up to the statement to use whatever defaults it wants (According to
		// specs this should be forward-only type and read-only concurrency,
		// but it's safer not to assume anything.)
		PreparedStatement preparedStatement = null;
		if (resultSetType == -1) {
			preparedStatement = conn.prepareStatement(sql);
		} else {
			preparedStatement = conn.prepareStatement(sql, resultSetType, resultSetConcurrency);
		}

		// Bind the values to the prepared statment.
		Object value = null;
		int parameterIndex = 1;
		for (Iterator iter = valuesToBind.iterator(); iter.hasNext();) {
			value = iter.next();
			if (logger.isDebugEnabled()) {
				logger.debug("calling setObject("+parameterIndex+","+value+")");
				if (value!=null) {
					logger.debug("value class is " + value.getClass().getName());
				}
			}
			preparedStatement.setObject(parameterIndex++,value);
		}

		// Finally return it.
		return preparedStatement;

	}

	/**
	 * <p>Populates this with sql loaded from the specified resource. If there
	 * already exists sql for keys in the specified resource, the sql will be
	 * updated.
	 *
	 * <p>The resource is loaded using the supplied Class's getResourceAsStream()
	 * method. If the name is prefixed with a "/" then the resource is absolute,
	 * otherwise it is relative to the class's package.
	 *
	 * <p>The resource must contain xml data in the following format.
	 *
	 * <pre>
	 * &lt;sql-queries&gt;
	 *     &lt;sql key="selectCustomer"&gt;
	 * 	       Select * from CUSTOMER Where CUSTOMER_ID = :customerId
	 *     &lt;/sql&gt;
	 * &lt;sql-queries&gt;
	 * </pre>
	 *
	 * @param resourceName - the name of the resource to load from.
	 * @param clazz - used to load the resource.
	 */
	public void addSql(
			String resourceName,
			Class clazz)
			throws StatementPreparerException {

		InputStream is = null;

		try {

			// Get the resource input stream from the specified class
			is = clazz.getResourceAsStream(resourceName);
			if (is==null) {
				logger.debug("could not get resource=" + resourceName + " using class=" + clazz);
				throw new StatementPreparerException(
						StatementPreparerException.MISSING_RESOURCE, resourceName, null);
			}

			// Parse the xml resource file using struts digester. See also:
			// http://jakarta.apache.org/commons/digester/api/org/apache/commons/digester/package-summary.html
			Digester digester = new Digester();
			digester.push(sqlMap);
			digester.addCallMethod("sql-queries/sql", "put", 2);
			digester.addCallParam("sql-queries/sql", 0, "key");
			digester.addCallParam("sql-queries/sql", 1);
			digester.parse(is);

			// Success! The map now contains all of the sql queries

		} catch (IOException ioe) {
			// Log and throw up as a StatementPreparerException
			logger.debug("io exception reading resource", ioe);
			throw new StatementPreparerException(
					StatementPreparerException.ERROR_READING_RESOURCE, resourceName, ioe);
		} catch (SAXException se) {
			// Log and throw up as a StatementPreparerException
			logger.debug("sax exception reading resource", se);
			throw new StatementPreparerException(
					StatementPreparerException.ERROR_READING_RESOURCE, resourceName, se);
		} finally {
			// Always try and close the input stream
			if (is!=null) {
				try {
					is.close();
				} catch (Exception e) {
					// do nothing..
				}
			}
		}

	}

}
