<#-- START TEMPLATE HEADER -->
<#-- category=generic -->
<#-- type=application -->
<#-- version=1.0 -->
<#-- fileNameTemplate=TemplateUtils.java -->
<#-- relativeOutputDirectory=src -->
<#-- relativePackage=util.template -->
<#-- END TEMPLATE HEADER -->
/*
 * Copyright (c) 2004-2005 Code Canvas Technologies Pty Ltd. All Rights Reserved.
 *
 * Modification and redistribution of this file is permitted subject to the
 * terms and conditions of the RapidJ End User License Agreement:
 * http://www.codecanvas.com.au/rapidj/license.php
 *
 * {{$RapidJ$}}
 */

/*
${f.formatLines(project.copyright, ' * ', 80)}
 */
package ${project.rootPackage}.util.template;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.util.Map;

import org.apache.log4j.Logger;

import freemarker.template.Configuration;
import freemarker.template.Template;

/**
 * TemplateUtils provides methods for evaluating a Freemarker template.
 */
public class TemplateUtils {

	/**
	 * Log4j Logger for this class
	 */
	public static final Logger logger = Logger.getLogger(TemplateUtils.class);

	/**
	 * Lazy loaded FreeMarker Configuration object. It should not be accessed
	 * directly but rather through the getCfg() method.
	 */
	private static Configuration cfg = null;

	/**
	 * Generate using the specified context and template snippet.
	 *
	 * @param templateSnippet - a String containing template code
	 * @param context - the context to use for the generate
	 * @return the generated content as a String
	 * @exception IOException
	 * @exception TemplateException
	 */
	public static String process(Map context, String templateSnippet) throws TemplateException {

		// Create a reader from the templateSnippet
		StringReader reader = new StringReader(templateSnippet);

		// Create a new freemarker template using reader
		Template fmTemplate = null;
		try {

			fmTemplate = new Template("snippet", reader, getCfg());

		} catch (IOException ioe) {

			// Wrap in a TemplateException
			logger.error("ioexception creating template", ioe);
			throw new TemplateException(
					TemplateException.PROCESSING_ERROR, null, ioe);

		}

		// Process the template
		String result = processTemplate(context, fmTemplate);

		// Return the result
		return result;

	}

	/**
	 * This method processes the template with the specified context.
	 *
	 * @param context - the context to use with the generate
	 * @param fmTemplate - the freemarker template
	 */
	private static String processTemplate(Map context, Template fmTemplate) throws TemplateException {

		OutputStreamWriter osw = null;
		ByteArrayOutputStream baos = null;

		logger.debug("processing... ");

		try {

			// Create the file output stream
			baos = new ByteArrayOutputStream();

			// Bridge the gap between output stream and a writer
			osw = new OutputStreamWriter(baos);

			// And merge the template with the context
			fmTemplate.process(context, osw);

			// Make sure all data is flushed
			osw.flush();

			// Return the processed Create the GeneratedContent
			String result = new String(baos.toByteArray());

			logger.debug("success!");

			return result;

		} catch (IOException ioe) {

			// Wrap in a TemplateException
			logger.error("ioexception processing template", ioe);
			throw new TemplateException(
					TemplateException.PROCESSING_ERROR, null, ioe);

		} catch (freemarker.template.TemplateException te) {

			// Wrap in a TemplateException
			logger.error("free marker template exception processing template", te);
			throw new TemplateException(
					TemplateException.PROCESSING_ERROR, null, te);

		} finally {
			if (osw!=null) {
				try {
					osw.close();
				} catch (Exception e) {}
			}
		}
	}


	/**
	 * Returns a lazy loaded reference to the Configuration
	 *
	 * @return
	 */
	private static Configuration getCfg() {
		if (cfg==null) {
			// Create a new freemarker configuration object
			cfg = new Configuration();
			// Set the delay to 0 to always gets the latest version of the
			// template from disk.
			cfg.setTemplateUpdateDelay(0);
		}
		return cfg;
	}

}
